import sys, os, shutil
import re
from collections import OrderedDict
import json
import numpy as np
from skimage.external import tifffile
from scipy.ndimage.interpolation import shift
from ncempy.edstomo.CharacteristicEmission import GetFluorescenceLineEnergy
from ncempy.io.emd import fileEMD

def BinEDSSpatialDimensions(Arr, Binning=4):
    if len(Arr.shape) != 3:
        print('Input stack needs to have a shape like (tilt, x, y).')
        return Arr

    if ((Arr.shape[1] % Binning) != 0) or ((Arr.shape[2] % Binning) != 0):
        print('Input stack needs to have dimensions that are evenly divided by the binning factor.')
        return Arr

    # Create new axes with a length of Binning and fold the array into them.
    ArrNew = Arr.reshape((  Arr.shape[0],                                   # Tilt dimension untouched
                            Binning,int(Arr.shape[1]/Binning),              # Spatial dimension binned
                            Binning,int(Arr.shape[2]/Binning)), order='F')   # Spatial dimension binned
    # Sum along the binning directions.
    ArrNew = np.sum(np.sum(ArrNew, axis=3), axis=1)
    # print('Old shape: ' + str(Arr.shape))
    # print('New shape: ' + str(ArrNew.shape))
    return ArrNew

def ExtractSignalsFromEMD(InputEMD=None, SignalNames=['HAADF', 'Mg_K', 'Fe_Ka'], Binning=4):
    ''' Read in a set of Bruker bcf files containing EDS acquisitions.

    Parameters:
        InputEMD (str): Name of EMD file containing the tilt stacks.

        SignalNames (list of str): A list of signals that should be extracted from the Bruker files.
            Valid values include:
            HAADF: For the HAADF signal
            Element_Line: e.g. Fe_Ka for a weighted sum of Fe_Ka1 and Ka2, or Fe_Ka1 for just that line, or Mg_K for a weighted sum of all Mg-K lines.  There is no brehmsstrahlung removal.
            eV1-eV2: Start and stop energies in eV.  The signal will be the sum of all energies over the range [eV1, eV2).

        Binning (int): How much binning to do on EDS signals to reduce noise.  1 means no binning.  2 means each output voxel is 2x2x2 input voxels.  HAADF signals are not rebinned as they are usually not noisy.

    Returns:
        SignalDict (OrderedDict of np.ndarray): Dictionary with names matching SignalNames and dimensions of (tilt, x, y).
        Tilts (list of floats): A list of tilt angles.

    '''

    # for s in SignalNames:
    #     if '_' in s:
    #         El, Line = s.split('_')
    #         print(GetFluorescenceLineEnergy(El, Series=Line[0], Line=Line))
    
    # Make an ordered dictionary with one entry for each signal.
    SignalDict = OrderedDict()
    for n in SignalNames:
        SignalDict[n] = []

    # Open the emd file that has our data.
    EMD = fileEMD(InputEMD, readonly=True)

    # We will compute the FWHM of peaks using the Mn-Ka reported in the EMD.
    EnergyResolutionMnKa = EMD.microscope.attrs['MnKaResolution[eV]']
    Mn_Ka_Energy = GetFluorescenceLineEnergy('Mn', Series='K', Line='Ka')
    K = EnergyResolutionMnKa / np.sqrt(Mn_Ka_Energy)
    print('Energy resolution of Mn-Ka is: ' + str(EnergyResolutionMnKa) + ' eV.')
    print('Assumed FWHM of peaks will be (%g*sqrt(E))/2., hence at Mn-Ka: %g eV.' % (K, K*np.sqrt(Mn_Ka_Energy)/2))

    # Get links to the data we'll need from the EMD.
    HAADF, HAADF_dims = EMD.get_emdgroup(EMD.data['HAADF_TiltStack'])
    EDS, EDS_dims  = EMD.get_emdgroup(EMD.data['EDS_TiltStack'])
    Tilts = HAADF_dims[0][0]
    print('HAADF dimensions are (%d, %d).'%(len(HAADF_dims[1][0]), len(HAADF_dims[1][0])))

    # Calculate the rebinning size for the EDS data.
    rebinsize_m = int(len(EDS_dims[1][0])/Binning)
    rebinsize_n = int(len(EDS_dims[2][0])/Binning)
    print('Binning is %d so rebinned EDS cubes will have spatial dimension (%d, %d).' % (Binning, rebinsize_m, rebinsize_n))

    for sig in SignalDict.keys():
        print(sig, end='')
        if sig == 'HAADF':
            SignalDict['HAADF'] = HAADF[:].astype("float32") # The HAADF doesn't get rebinned.
            print('')

        if '_' in sig:
            # This is a fluorescence line.  
            El, Line = sig.split('_')
            CenterEnergy = GetFluorescenceLineEnergy(El, Series=Line[0], Line=Line)
            if CenterEnergy is None:
                print('Unrecognized fluorescence line: %s-%s, ignoring this signal.'%(El,Line))
                continue
            LowEnergy = CenterEnergy - K*np.sqrt(CenterEnergy)/2
            HighEnergy = CenterEnergy + K*np.sqrt(CenterEnergy)/2
            LowEnergyIndex = np.argmin((EDS_dims[3][0] - LowEnergy)**2)
            HighEnergyIndex = np.argmin((EDS_dims[3][0] - HighEnergy)**2)
            print(', %g-%g eV window, energy bins: %d-%d.'% (LowEnergy, HighEnergy, LowEnergyIndex, HighEnergyIndex))
            Cube = np.sum(EDS[:, :, :, LowEnergyIndex:HighEnergyIndex+1], axis=-1)
            SignalDict[sig] = BinEDSSpatialDimensions(Cube, Binning)
            # fluor[0].rebin((rebinsize_m,rebinsize_n)).data.copy().astype("float32"))
        
        if '-' in sig:
            print('Energy range signals not implemented yet.')
    
    print('Signals Extracted.')

    # Turn those signal readouts into 3D numpy arrays (tilt, x, y).
    for k, v in SignalDict.items():
        npStack = np.array(v)
        SignalDict[k] = npStack

    return SignalDict, Tilts
    
def NormalizeSignals(SignalDict, Tilts, NormalizationSignalName=None, NormalizationImageFraction=1):
    ''' 
    Parameters:
        SignalDict (OrderedDict of np.ndarray): Dictionary with signal stacks.  Each entry is one tilt sequence with dimension (tilt, x, y).  The key is the name of the signal, e.g. 'Fe_Ka' or 'HAADF'.

        Tilts (list of floats): A list of tilt angles.

        NormalizationSignalName (str): Signal name to use for normalizing signal.  Preferably a higher energy but still high signal line.  Note, using HAADF for normalization produced bad results since the HAADF signal is not shadowed by the tomography holder.  None means don't normalize, 'Theory' means to normalize using a theoretical curve.

        NormalizationImageFraction (float): Fraction of frame (from center) to use to produce a normalization curve.  1 means to use the entire frame.  0.5 means to use the center half of the frame (0.25 - 0.75 of the frame width and height).

    Returns:
        SignalDict (OrderedDict of np.ndarray): Same as passed in but normalized.
        NormCurve (np.ndarray): Values to multiply by each tilt image in order to normalize the signal.
    '''

    # Normalize the spectra.
    NormCurve = np.ones(len(Tilts)) # Default to no normalization -- we're just multiplying by identity.
    if NormalizationSignalName is not None:
        if NormalizationSignalName == 'Theory':
            print('Theoretical normalization curve not implemented yet.')
        elif NormalizationSignalName == 'Independent':
            print('Producing normalization curve for each signal independently.')
            for k, v in SignalDict.items():
                NormCurve = GetNormalizationCurve(SignalDict[k], Tilts, NormalizationImageFraction)
                v = v * NormCurve[:,None,None]
                SignalDict[k] = v
        else:
            print('Producing normalization curve for all signals based on '+NormalizationSignalName+'.')
            NormCurve = GetNormalizationCurve(SignalDict[NormalizationSignalName], Tilts, NormalizationImageFraction)
            for k, v in SignalDict.items():
                v = v * NormCurve[:,None,None]
                SignalDict[k] = v

    return SignalDict, NormCurve

def GetNormalizationCurve(Signal, Tilts, ImageFraction=1):
    ''' Given a signal, finds the normalization curve to correct for the fact that sample thickness changes,
    and EDS detector shadowing changes as a function of tilt angle.

    Parameters:
        Signal (np.ndarray): Shape of (Tilts, x, y).  Signal from which to derive a normalization curve.

        Tilts: (np.ndarray): Stage tilts for the tomographic acquisition.

        ImageFraction (float): Fraction of the image to use to produce a norm curve.  Often higher tilt images see additional material coming into view which negates the values of the normalization curve.  1 means use the whole frame and would be valid for an isolated object entirely in the FOV.  0.5 means use only the range in x and y from 0.25-0.75 of the width and height of the frame.  Must be in the range 0 < ImageFraction <= 1.

    Returns:
        NormCurve (np.ndarray): Values to multiply by each tilt image in order to normalize the signal.
    '''

    # Convert to float
    Signal = Signal.astype(float)

    # Figure out the portion out of the center of the image to use to determine the curve.
    xmin = int(np.floor(Signal.shape[1] * ((1/2)-ImageFraction/2)))
    xmax = int(np.ceil(Signal.shape[1] * ((1/2)+ImageFraction/2)))
    ymin = int(np.floor(Signal.shape[2] * ((1/2)-ImageFraction/2)))
    ymax = int(np.ceil(Signal.shape[2] * ((1/2)+ImageFraction/2)))
    # Crop the signal to only include the center portion.
    Signal = Signal[:, xmin:xmax, ymin:ymax]

    # Get the index of the stack with the least tilt.
    MinTilt = np.argmin(np.array(Tilts)**2)
    # Now sum up the minimum tilt frame (usually zero tilt.)
    CenterIntensity = np.sum(Signal[MinTilt,:,:])
 
    # The normalization curve will have one value for each tilt.
    NormCurve = np.zeros(len(Tilts))

    # Loop through each frame and integrate it to get intensity.
    for i in range(Signal.shape[0]):
        NormCurve[i] = np.sum(Signal[i,:,:])
 
    # Set this so the center intensity is 1.  Therefore, after normalization, the least tilt map is unchanged.
    NormCurve /= CenterIntensity

    # Right now higher intensities mean we counted more counts due to a decrease in detector shadowing.
    # We want to multiply each stack by the value of NormCurve, so we need inverse.
    NormCurve = 1/NormCurve

    return NormCurve

def ReadImageJTranslations(FileName, Tilts):
    ''' The ImageJ multistackreg plugin is one way to align stacks.  This method reads in the translations file written by ImageJ.

    Parameters:
        FileName (str): Name of the file that was created by ImageJ.
        Tilts: (np.ndarray): Stage tilts for the tomographic acquisition.

    Returns:
        np.ndarray with shape (Tilts, 2) containing an x and y translation for each tilt.
    '''

    with open(FileName, 'r') as f:
        s = f.read()
        
    translations = re.findall(pattern='([\d\.]*)\t([\d\.]*)', string=s)

    # There are 6 lines in each translation from the output file.  
    # The zeroth is the new position, the third is the original position.
    # The other lines are for rotations which we are not using.
    # Example:
    # [('236.81374221505683', '288.85170101351565'),
    #  ('0.0', '0.0'),
    #  ('0.0', '0.0'),
    #  ('256.0', '256.0'),
    #  ('0.0', '0.0'),
    #  ('0.0', '0.0'),
    #  ('246.23227371186195', '229.14170151652533'),
    #  ('0.0', '0.0'),
    #  ('0.0', '0.0'),
    #  ('256.0', '256.0'),
    #  ('0.0', '0.0'),
    #  ('0.0', '0.0'),
    #  ('264.2877416986045', '281.32142526192746'),
    #  ('0.0', '0.0'),
    #  ('0.0', '0.0'),
    #  ('256.0', '256.0'),
    #  ('0.0', '0.0'),
    #  ('0.0', '0.0'),
    #  ...
     

    # There should be as many translations as there are images-1 because we are registering to the first image.
    assert(len(translations)/6 == len(Tilts)-1)

    offsets = []
    offsets.append(np.array((0,0))) # MultiStackReg will ignore the first image since others are registered relative to it.
    # Loop through all the translation info, image by image (6 lines at a time).
    for i in range(0, len(translations), 6):
        NewPos = translations[i] # Image shifted to here
        OrigPos = translations[i+3] # Image started here
        # How much to shift:
        offset = -np.array((float(NewPos[1]) - float(OrigPos[1]), float(NewPos[0]) - float(OrigPos[0])))
        if len(offsets) > 0:
            offset += offsets[-1]
        offsets.append(offset)

    offsets = np.array(offsets)
    print(offsets)
    return offsets

def ReadTomVizTranslations(FileName, Tilts):
    ''' TomViz has a capability for aligning stacks.  The manual method works very well because EDS tomographs usually only have a small number of frames.  This function reads in alignements generated by the TomViz manual method and saved to a json file.

    Parameters:
        FileName (str): Name of the file that was created by TomViz.
        Tilts: (np.ndarray): Stage tilts for the tomographic acquisition.

    Returns:
        np.ndarray with shape (Tilts, 2) containing an x and y translation for each tilt.
    '''

    with open(FileName, 'r') as f:
        s = json.load(f)
    s = np.array(s)
    # The coordinates in tomviz are different.  X and Y are swapped, and the sign of one axis is different.
    s = np.roll(s, 1, axis=1)
    s[:,0] = -s[:,0]
    return s

def ApplyTranslations(SignalDict, Translations, AlignSignalName):
    ''' Apply translations to all the images in all the signals.

    Parameters:
        SignalDict (OrderedDict of np.ndarray): Dictionary with signal stacks.  Each entry is one tilt sequence with dimension (tilt, x, y).  The key is the name of the signal, e.g. 'Fe_Ka' or 'HAADF'.

        Translations (np.ndarray with shape (Tilts, 2)): An x and y translation for each tilt.
        
        AlignSignalName (str): Name of the signal used to align the stack.  e.g. 'Fe_Ka' or 'HAADF'

    Returns:
        AlignedSignalDict: SignalDict after applying translations to each image in each signal.

    '''

    AlignedSignalDict = OrderedDict()

    # Alignment is done with one signal.  Usually this is HAADF but sometimes it is an EDS channel.
    # Since the binning is not the same, we have to scale the alignment to the other channels.
    AlignSignalShape = SignalDict[AlignSignalName].shape

    for SignalName, Sig in SignalDict.items():
        print('Aligning', SignalName)
        SigAlign = []

        # Calculate how much to translate this image based on it's shape relative to the signal used for alignment.
        TranslationScale = np.ones(2)
        TranslationScale[0] = Sig.shape[1]/AlignSignalShape[1]
        TranslationScale[1] = Sig.shape[2]/AlignSignalShape[2]
        print('TranslationScale: '+str(TranslationScale))

        # Apply the shifts to every image in this signal stack.
        for i in range(len(Translations)):
            ThisShift = np.round(Translations[i]*TranslationScale).astype(int)
            SigAlign.append(shift(Sig[i], ThisShift))

        # Turn the individual shifted images back into a 3D numpy array with dimension (tilts,x,y).
        SigAlign = np.array(SigAlign)

        AlignedSignalDict[SignalName] = SigAlign
    return AlignedSignalDict

def WriteSignalsToTIFFs(OutputDirectory, SignalDict):
    ''' Write the signals we have to tiff stacks so that we can do alignment, GENFIRE, etc. on them.  Also write auxiliary files that are needed by these processes.

    Parameters:
        OutputDirectory (str): Name of directory to place the tif files.

        SignalDict (OrderedDict of np.ndarray): Dictionary with signal stacks.  Each entry is one tilt sequence with dimension (tilt, x, y).  The key is the name of the signal, e.g. 'Fe_Ka' or 'HAADF'.

    Returns:
        Files written.
    '''

    # Ensure the output directory exists and is ready for us to write our output.
    if not os.path.exists(OutputDirectory):
        print('Creating directory: '+OutputDirectory)
        os.makedirs(OutputDirectory)

    # Write out all the stacks as tiff files.
    for k, v in SignalDict.items():
        # Reformat the stacks to have the correct shape and bit depth to write to tifs.
        npStack = v.astype(float)

        # Things need to fit in 16 bits.
        if np.max(npStack) > 2**16-1:
            npStack /= np.max(npStack)
            npStack *= 2**16-1
        
        # Now convert to uint16 for 16 bit unsigned tifs.
        npStack = npStack.round().astype('uint16')

        # Finally we get to save the actual tif.
        print('Writing '+str(k)+'.tif')
        tifffile.imsave(os.path.join(OutputDirectory,k+'.tif'), npStack)

def ReadSignalsFromTIFFs(InputDirectory, SignalNames=['HAADF', 'Mg_Ka', 'Fe_Ka']):
    ''' Read from TIFF files each of the requested signals.
    Parameters:
        InputDirectory (str): Name of directory containing the bcf files.

        SignalNames (list of str): A list of signals that should be extracted from the Bruker files.
            Valid values include:
            HAADF: For the HAADF signal
            Element_edge: e.g. Fe_Ka for a specific line.  There is no brehmsstrahlung removal.
            eV1-eV2: Start and stop energies in eV.  The signal will be the sum of all energies over the range [eV1, eV2).

    Returns:
        SignalDict (OrderedDict of np.ndarray): Dictionary with signal stacks.  Each entry is one tilt sequence with dimension (tilt, x, y).  The key is the name of the signal, e.g. 'Fe_Ka' or 'HAADF'.


    '''
    SignalDict = OrderedDict()
    for SigName in SignalNames:
        npStack = tifffile.imread(os.path.join(InputDirectory, SigName+'.tif')).astype(float)
        SignalDict[SigName] = npStack
    return SignalDict

def WriteMetaDataFiles(OutputDirectory, Tilts, NormCurve=None, NormalizationSignalName=None):
    ''' Writes out meta data obtained while processing the inputs which may be useful.
    Parameters:
        OutputDirectory (str): Name of directory to place the tif files.

        Tilts (list of floats): A list of tilt angles.

        NormCurve (np.ndarray): Values to multiply by each tilt image in order to normalize the signal.

    Returns:
        Files written.
    '''
    # Make the tilts file
    with open(os.path.join(OutputDirectory,'tilts.txt'), 'w') as f:
        print('Writing tilts file.')
        for n in Tilts:
            f.write('0.0 %0.1f 0.0\n'%(n))
    # Write out the normalization curve.
    if NormCurve is not None:
        np.savetxt(os.path.join(OutputDirectory, 'Normalization_Curve.txt'), NormCurve.T, header=NormalizationSignalName)

def ReadMetaDataFiles(InputDirectory):
    ''' Writes out meta data obtained while processing the inputs which may be useful.
    Parameters:
        InputDirectory (str): Name of directory that has the metadata files.

    Returns:
        Tilts (list of floats): A list of tilt angles.
        NormCurve (np.ndarray): Values to multiply by each tilt image in order to normalize the signal.
        NormalizationSignalName (str): Name of the signal that was used to normalize.
    '''
    # Read the tilts file
    print('Reading tilts file.')
    TiltsTemp = np.genfromtxt(os.path.join(InputDirectory,'tilts.txt'))
    Tilts = TiltsTemp[:,1]

    # Read the normalization curve.
    print('Reading normalization curve.')
    NormCurve = np.genfromtxt(os.path.join(InputDirectory, 'Normalization_Curve.txt'), names=True)
    NormalizationSignalName = NormCurve.dtype.names[0]

    return Tilts, NormCurve, NormalizationSignalName

def WriteSignalsToGENFIRE(OutputDirectory, SignalDict, Tilts, GENFIRETemplateDir=''):
    ''' Write npy files that GENFIRE can read and ensure all the metadata it needs are present.

    Parameters:
        OutputDirectory (str): Name of directory to place the tif files.

        SignalDict (OrderedDict of np.ndarray): Dictionary with signal stacks.  Each entry is one tilt sequence with dimension (tilt, x, y).  The key is the name of the signal, e.g. 'Fe_Ka' or 'HAADF'.

    Returns:
        A npy file for each signal.
        A script file for each signal to run GENFIRE using DoGenfire.py.
        DoGenfire.py is present in the output directory.
        A script file to run all the script files.
    '''

    # For now, rotation axis is hard coded.  We actually need to address the full rotation issue later.
    # Right now, I just assume the user did a perfect job of aligning the EDS stack around the stage alpha axis.  >;-)
    RotationAxisVertical = True # If false, then the rotation axis is going horizontally across the image.

    # Right now, we only support unix-like systems.  This currently runs on OS-X and Scientific Linux 7.  But it could easily be extended to other OS options. 

    # Read in the template for the script file to run GENFIRE.
    with open(os.path.join(os.path.dirname(__file__), 'slurm_template.sh'), 'r') as f:
        slurmscripttemplate = f.read()

    # Make sure the output directory has the DoGenfire.py file.
    shutil.copyfile(os.path.join(os.path.dirname(__file__), 'DoGenfire.py'), os.path.join(OutputDirectory, 'DoGenfire.py'))

    # Start a script file that will run all the script files.
    runall = open(os.path.join(OutputDirectory,'runall.sh'), 'w')

    # Write a npy file for each signal.
    for SigName, Sig in SignalDict.items():

        # Swap the numpy axes to be how GENFire expects it and save npy files.
        Sig = np.swapaxes(np.swapaxes(Sig, 0,1), 1,2)
        if(RotationAxisVertical==True):
            Sig = np.swapaxes(Sig, 0,1)
        print('Writing '+str(SigName)+'_aligned.npy, ', end='')
        np.save(os.path.join(OutputDirectory,SigName+'_aligned.npy'), np.array(Sig))

        # Write out a script file which will run GENFIRE for this MRC.
        # HAADF images are usually too big to do the full genfire.  We have avoid using oversampling so they fit in memory.
        if SigName == 'HAADF':
            osetting = '-o 1'
        else:
            osetting = ''

        slurmscript = slurmscripttemplate.replace('$JOBNAME', 'GENFIRE_'+SigName)
        slurmscript = slurmscript.replace('$GENFIRESTRING', SigName+' '+osetting+'\n')
        with open(os.path.join(OutputDirectory,SigName+'_slurm.sh'), 'w') as f:
            print('Writing '+SigName+'_slurm.sh')
            f.write(slurmscript)
        
        runall.write('sbatch '+SigName+'_slurm.sh\n')
        runall.write('sleep 3\n')
    print('Writing runall.sh')
    runall.close()

if __name__ == '__main__':
    print('------------------------------ Begin preprocess.py test ------------------------------')


    # InputDirectory = os.path.join('..', 'data', 'L2083-K-4-1', 'Input')
    # OutputDirectory = os.path.join('..', 'data', 'L2083-K-4-1', 'TestOutput', 'Unaligned')
    # AlignedDirectory = os.path.join('..', 'data', 'L2083-K-4-1', 'TestOutput', 'Aligned')
    OutputDirectory = os.path.join('..', 'data', 'L2083-K-4-1', 'TestOutput')
    SignalNames = ['HAADF', 'Al_Ka', 'C_Ka', 'Ca_Ka', 'Cr_Ka', 'Fe_Ka', 'Ga_Ka', 'Mg_Ka', 'Na_Ka', 'Ni_Ka', 'O_Ka', 'P_Ka', 'Pt_La', 'S_Ka', 'Si_Ka']
    EMDFileName = os.path.join('..', 'data', 'L2083-K-4-1', 'ConvertBrukerToEMD', 'Output', 'TomoData.emd')

    Signals, Tilts = ExtractSignalsFromEMD(EMDFileName, SignalNames=['HAADF', 'Fe_Ka', 'Si_K', 'Mg_K'])
    Signals, NormCurve = NormalizeSignals(Signals, Tilts, NormalizationSignalName='Si_K', NormalizationImageFraction=0.5)
    # WriteSignalsToTIFFs(OutputDirectory, Signals)

    # # Read in the data and process it up to the point where we need to do stack alignment in an external program.
    # if False:
    #     Signals, Tilts = ExtractRawSignalsFromBrukerSequence(InputDirectory=InputDirectory, SignalNames=SignalNames)
    #     print('Signal Names: '+str(Signals.keys()))
    #     print('HAADF signal shape: '+str(Signals["HAADF"].shape))
    #     print('Fe_Ka signal shape: '+str(Signals["Fe_Ka"].shape))
    #     print('Tilts: '+str(Tilts))
    #     Signals, NormCurve = NormalizeSignals(Signals, Tilts, NormalizationSignalName='Fe_Ka', NormalizationImageFraction=0.5)
    #     print(NormCurve)
    #     WriteSignalsToTIFFs(OutputDirectory, Signals)
    #     WriteMetaDataFiles(OutputDirectory, Tilts, NormCurve, NormalizationSignalName='Fe_Ka')
    #     input('Please do the stack alignment in ImageJ with the Multistackreg plugin now.  Press any key to continue...')

    # # At this point you should align the stacks.

    # # Apply the alignment file to generate aligned stacks and produce input into GENFIRE.
    # if True:
    #     SignalDict = ReadSignalsFromTIFFs(OutputDirectory, SignalNames=SignalNames)
    #     Tilts, NormCurve, NormalizationSignalName = ReadMetaDataFiles(OutputDirectory)
    #     Translations = ReadTomVizTranslations(os.path.join(OutputDirectory, 'TomVizAlignments.json'), Tilts)
    #     # Translations = ReadImageJTranslations(os.path.join(OutputDirectory, 'TransformationMatrices.txt'), Tilts)
    #     AlignedSignals = ApplyTranslations(SignalDict, Translations, 'HAADF')
    #     WriteSignalsToTIFFs(AlignedDirectory, AlignedSignals)
    #     WriteMetaDataFiles(AlignedDirectory, Tilts, NormCurve, NormalizationSignalName='Fe_Ka')
    #     WriteSignalsToGENFIRE(AlignedDirectory, AlignedSignals, Tilts)
    #     print('Now run GENFIRE using the (SignalName)_aligned.sh scripts.')

    # # At this point run GENFIRE.

    print('------------------------------ End preprocess.py test ------------------------------')



import numpy as np

from ncempy.eval import multicorr


def stack_align(stack, align_type='static', **kwargs):
    """ Align a series of images by correlation using multicorr. All images are aligned to the start image or as
    dynamically starting with the start image and then each successive image. keyword arguments are passed to
    multicorr. Shifting is done in Fourier space which is very accurate, but wraps edges.

    Parameters
    ----------
        stack : ndarray, 3D,
            Stack of images to align. Shape [num, Y, X]

        align_type: str
            static or dynamic alignment. Static aligns all images to the first image. Dynamic aligns
            each image to the previous image starting with the first image

    Returns
    -------
        : tuple, aligned stack, shifts
            A tuple containing the aligned images as a 3D ndarray of shape
            [num, Y, X] and shifts as a 2D ndarray of shape [num, 2]

    """

    if align_type is not 'static' and align_type is not 'dynamic':
        raise KeyError('Incorrect align type. Must be static or dynamic')

    # Pre-allocate the arrays
    aligned = np.zeros_like(stack)  # shifted data array
    shifts = np.zeros((stack.shape[0], 2))  # the applied shifts

    aligned[0, :, :] = stack[0, :, :]

    # Align positive angles
    ref_fft = np.fft.fft2(stack[0, :, :])
    ref_sh = np.zeros((2,))
    for ii in range(1, stack.shape[0]):
        cur_fft = np.fft.fft2(stack[ii, :, :])
        sh = multicorr.multicorr(cur_fft, ref_fft, **kwargs)
        sh += ref_sh
        image_shifted = np.real(np.fft.ifft2(multicorr.imageShifter(cur_fft, sh)))
        aligned[ii, :, :] = image_shifted
        shifts[ii, :] = sh

        if align_type is 'dynamic':
            ref_fft = cur_fft
            ref_sh = sh

    return aligned, shifts



import sys, os, shutil
from collections import OrderedDict
import numpy as np
import genfire

CropBounds = {'xmin': 0.10, 'xmax': 0.90,
              'ymin': 0.10, 'ymax': 0.90,
              'zmin': 0.10, 'zmax': 0.90}

def ReadGENFIRESignals(GENFIREDirectory, SignalNames):
    SignalDict = OrderedDict()

    for SigName in SignalNames:
        Sig = genfire.fileio.readMRC(SigName+'_reconstruction.mrc')
        SignalDict[SigName] = Sig

def SquareCropSignal(Signal, CropBounds = None):
    # x axis
    Signal[:,:, :int(CropBounds['xmin']*Signal.shape[2])] = 0 # Min for axis
    Signal[:,:, int(CropBounds['xmax']*Signal.shape[2]):] = 0 # Max for axis.

    # y axis
    Signal[:, :int(CropBounds['ymin']*Signal.shape[1]), :] = 0 # Min for axis
    Signal[:, int(CropBounds['ymax']*Signal.shape[1]):, :] = 0 # Max for axis.
    
    # z axis
    Signal[:int(CropBounds['zmin']*Signal.shape[0]), :, :] = 0 # Min for axis
    Signal[int(CropBounds['zmax']*Signal.shape[0]):, :, :] = 0 # Max for axis.

    return Signal



"""
This module provides an interface to the Berkeley EMD file format.

See https://emdatasets.com/ for more details.

* It fully supports v0.2 for reading and writing

Note
----
General users:
    Use the simplified emd.emdReader() function to load the data and meta
    data as a python dictionary.

Advanced users and developers:
    Access the file internals through the emd.fileDEMD() class.

"""

from pathlib import Path
import datetime

import numpy as np
import h5py


class fileEMD:
    """Class to represent Berkelely EMD files.

    Implemented for spec 0.2 using the recommended layout for metadata.

    Attributes
    ----------

    file_hdl : h5py.File
        The h5py file handle which provides direct access to the underlying h5py file structure.
    version : tuple
        A 2-tuple of the major and minor EMD version as ints.
    data : h5py.Group
        A link to the data group in the EMD file.
    microscope : h5py.Group
        A link to the microsocpe EMD group with metadata.
    sample : h5py.Group
        A link to the sample EMD group with metadata.
    user : h5py.Group
        A link to the user EMD group with metadata.
    comments : h5py.Group
        A link to the comments in this EMD file. A time stamp is included with the comment.
    list_emds : list
        A list of the EMD groups in the file.

    Examples
    --------
    Open an Berkeley EMD file using the *advanced* interface. See the emdReader function for a more convenient
    way to load the data. We show how to list the available data sets, load a 3D data set and plot the first image.

        >>> from ncempy.io import emd
        >>> with emd.fileEMD('filename.emd') as emd1:
        >>>     [print(dataGroup.name) for dataGroup in emd1.list_emds] # print all available EMD datasets
        >>>     data1, dims1 = emd1.get_emdgroup(0) # load the first full data array and dimension information
    """

    def __init__(self, filename, readonly=True):
        """Init opening/creating the file.

        Parameters
        ----------
        filename : str or pathlib.Path
            Path to the EMD file.
        readonly : bool, default True
            Set to False to allow writing to the file.

        """

        # necessary declarations in case something goes bad
        self.file_hdl = None

        # convenience handles to access the data in the emd file, everything can as well be accessed using the file_hdl
        self.version = None
        self.data = None
        self.microscope = None
        self.sample = None
        self.user = None
        self.comments = None
        self.list_emds = []  # list of HDF5 groups with emd_data_type type 1

        # check filename type, change to string
        if isinstance(filename, str):
            pass
        elif isinstance(filename, Path):
            filename = str(filename)
        else:
            raise TypeError('Filename is supposed to be a string or pathlib.Path')

        # try opening the file
        if readonly:
            try:
                self.file_hdl = h5py.File(filename, 'r')
            except:
                print('Error opening file for readonly: "{}"'.format(filename))
                raise
        else:
            try:
                self.file_hdl = h5py.File(filename, 'a')
            except:
                print('Error opening file for read/write: "{}"'.format(filename))
                raise

        # if we got a working file
        if self.file_hdl:
            # check version information
            if 'version_major' in self.file_hdl.attrs and 'version_minor' in self.file_hdl.attrs:
                # read version information
                self.version = (self.file_hdl.attrs['version_major'], self.file_hdl.attrs['version_minor'])
                # compare to implementation
                if not self.version == (0, 2):
                    print('WARNING: You are reading a version {}.{} EMD file, '
                          'this implementation assumes version 0.2!'.format(self.version[0], self.version[1]))
            else:
                # set version information
                if not readonly:
                    self.file_hdl.attrs['version_major'] = 0
                    self.file_hdl.attrs['version_minor'] = 2

            # check for data group
            if 'data' not in self.file_hdl:
                if not readonly:
                    self.data = self.file_hdl.create_group('data')
            else:
                self.data = self.file_hdl['data']

            # check for data group
            if 'microscope' not in self.file_hdl:
                if not readonly:
                    self.microscope = self.file_hdl.create_group('microscope')
            else:
                self.microscope = self.file_hdl['microscope']

            # check for data group
            if 'sample' not in self.file_hdl:
                if not readonly:
                    self.sample = self.file_hdl.create_group('sample')
            else:
                self.sample = self.file_hdl['sample']

            # check for data group
            if 'user' not in self.file_hdl:
                if not readonly:
                    self.user = self.file_hdl.create_group('user')
            else:
                self.user = self.file_hdl['user']

            # check for data group
            if 'comments' not in self.file_hdl:
                if not readonly:
                    self.comments = self.file_hdl.create_group('comments')
            else:
                self.comments = self.file_hdl['comments']

            # find emd_data_type groups in the file
            self.list_emds = self.find_emdgroups(self.file_hdl)

    def __del__(self):
        """Destructor for EMD file object.

        """
        # close the file
        # if(not self.file_hdl.closed):
        self.file_hdl.close()

    def __enter__(self):
        """Implement python's with statement

        """
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Implement python's with statement
        and close the file via __del__()
        """
        self.__del__()
        return None

    def find_emdgroups(self, parent):
        """Find all emd_data_type groups within the group parent and return a list of references to their HDF5 groups.

        Parameters
        ----------
            parent: h5py.Group
                Handle to the parent group.

        Returns
        -------
            : list
                A list of h5py.Group handles to children groups
                being emd_data_type groups.

        """

        emds0 = []

        # recursive function to run and retrieve groups with emd_group_type set to 1
        def proc_group(group, emds):
            # take a look at each item in the group
            for item in group:
                # check if group
                if group.get(item, getclass=True) == h5py.Group:
                    item = group.get(item)
                    # check if emd_group_type
                    if 'emd_group_type' in item.attrs and 'data' in item.keys():
                        if item.attrs['emd_group_type'] == 1:
                            emds.append(item)
                    # process subgroups
                    proc_group(item, emds)

        # run
        proc_group(parent, emds0)

        return emds0

    def get_emddims(self, group):
        """Get the emdtype dimensions saved in in group.

        Parameters
        ----------
            group: h5py.Group
                Reference to the emdtype HDF5 group.

        Returns
        -------
            : tuple
                List of dimension vectors plus labels and units.

        """
        # get the dims
        dims = []
        for ii in range(group['data'].ndim):
            dim = group['dim{}'.format(ii + 1)]
            # save them as (vector, name, units)

            if isinstance(dim.attrs['name'], np.ndarray):
                name = dim.attrs['name'][0]
            else:
                name = dim.attrs['name']

            if isinstance(dim.attrs['units'], np.ndarray):
                units = dim.attrs['units'][0]
            else:
                units = dim.attrs['units']

            # Handle bytes objects by decoding them to strings
            # If something goes wrong, the original attribute is left as-is
            try:
                if isinstance(name, bytes):
                    name = name.decode('utf-8')
                if isinstance(units, bytes):
                    units = units.decode('utf-8')
            except:
                pass

            dims.append((dim[:], name, units))

        dims = tuple(dims)
        return dims

    def get_emdgroup(self, group, memmap=False):
        """Get the emd data saved in the requested group.

        Parameters
        ----------
            group: h5py._hl.group.Group or int
                Reference to the HDF5 group to load. If int is used then the item corresponding to self.list_emds
                is loaded
            memmap: bool
                Return the data as a memmap instead of loading everything into memory

        Returns
        -------
            : tuple or None
                None or tuple containing:
                    : np.ndarray
                        The data of the emdtype group.
                    : list
                        List of [0] dimension vectors, [1] labels and [2] units.
        """

        # check input
        if not isinstance(group, h5py.Group):
            if isinstance(group, int):
                try:
                    group = self.list_emds[group]
                except IndexError:
                    print('group does not exist')
                    return
            else:
                raise TypeError('group needs to refer to a valid HDF5 group!')

        if 'emd_group_type' not in group.attrs:
            raise TypeError('group is not a emd_group_type group!')
        if not group.attrs['emd_group_type'] == 1:
            raise TypeError('group is not a emd_group_type group!')

        # retrieve data
        try:
            # get the data
            if memmap:
                data = group['data']
            else:
                data = group['data'][:]

            # get the dimensions.
            dims = self.get_emddims(group)

            return data, dims
        except:
            # if something goes wrong, return None
            print('Content of "{}" does not seem to be in emd specified shape'.format(group.name))

            return None

    def write_dim(self, label, dim, parent):
        """Auxiliary function to write a dim dataset to parent.

        Input is not checked for sanity, so handle exceptions in call.

        Parameters
        ----------
            label: str
                Label for dataset, usually dim1, dim2, dimN.
            dim: tuple
                Tuple containing (data, name, units).
            parent: h5py.Group
                HDF5 handle to parent group.

        Returns
        -------
            : h5py.Group
                HDF5 dataset handle referencing this dim.

        """

        try:
            dset = parent.create_dataset(label, data=dim[0])
            dset.attrs['name'] = np.string_(dim[1])
            dset.attrs['units'] = np.string_(dim[2])
        except:
            raise RuntimeError('Error during writing dim dataset')

        return dset

    def put_emdgroup(self, label, data, dims, parent=None, overwrite=False, **kwargs):
        """Put an emdtype dataset into the EMD file.

        Parameters
        ----------
            label: str
                Label for the emdtype group containing the dataset.
            data: np.ndarray
                Numpy array containing the data.
            dims: tuple
                Tuple containing the necessary dims as ((vec, name, units), (vec, name, units), ...)
            parent: h5py.Group or None
                Parent for the emdtype group, if None it will be written to /data.
            overwrite: bool
                Set to force overwriting entry in EMD file.
            **kwargs: various
                Keyword arguments to be passed to h5py.create_dataset(), e.g. for compression.

        Returns
        -------
            : h5py.Group or None
                Group referencing this emdtype dataset or None if failed.
        """

        # check input
        if not isinstance(label, str):
            raise TypeError('label needs to be string!')

        if not isinstance(data, np.ndarray):
            raise TypeError('data needs to be a numpy.ndarray!')

        try:
            assert len(dims) == len(data.shape)
            for i in range(len(dims)):
                assert len(dims[i]) == 3
                assert dims[i][0].shape[0] == data.shape[i]
        except:
            raise TypeError('Something wrong with the provided dims')

        # write stuff to HDF5

        # create group
        try:
            if parent:
                if label in parent:
                    if overwrite:
                        print('overwriting "{}" in "{}"'.format(label, parent.name))
                        del parent[label]
                    else:
                        print('"{}" already exists in "{}"'.format(label, parent.name))
                        raise RuntimeError('"{}" already exists in "{}"'.format(label, parent.name))
                grp = parent.create_group(label)

            else:
                if label in self.data:
                    if overwrite:
                        print('overwriting "{}" in "{}"'.format(label, self.data.name))
                        del self.data[label]
                    else:
                        print('"{}" already exists in "{}"'.format(label, self.data.name))
                        raise RuntimeError('"{}" already exists in "{}"'.format(label, self.data.name))

                grp = self.data.create_group(label)

            # add attribute
            grp.attrs['emd_group_type'] = 1

            # create dataset
            _ = grp.create_dataset('data', data=data, **kwargs)

            # create dim datasets
            for i in range(len(dims)):
                self.write_dim('dim{}'.format(i + 1), dims[i], grp)

            # update emds list
            self.list_emds = self.find_emdgroups(self.file_hdl)

            return grp

        except:
            print('Something went wrong trying to write the dataset.')

            return None

    def put_comment(self, msg, timestamp=None):
        """Create a comment in the EMD file.

        If timestamp already exists, the msg is appended to existing comment.

        Parameters
        ----------
            msg: str
                String of the message to save.
            timestamp: str/None
                Timestamp used as the key, defaults to the current UTC time.

        """

        # check input
        if not isinstance(msg, str):
            raise TypeError('msg needs to be a string!')

        # create timestamp if missing
        if not timestamp:
            timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S (UTC)')
        else:
            # try to convert given timestamp to string
            try:
                timestamp = str(timestamp)
            except:
                raise

        # write comment
        if timestamp in self.comments.attrs:
            # append to existing
            self.comments.attrs[timestamp] += np.string_('\n' + msg)

        else:
            # create new entry
            self.comments.attrs[timestamp] = np.string_(msg)
            
    def get_memmap(self, group):
        """ Get the emd group data as a memmap so that the data
        is not loaded into memory. Essentially calls get_emdgroup()
        with the keyword memmap keyword equals True.
        
        See get_emdgroup() for parameters and return values.
        
        
        """
        return self.get_emdgroup(group, memmap=True)
        

def defaultDims(data, pixel_size=None):
    """ A helper function that can generate a properly setup dim tuple
    with default values to allow quick writing of EMD files without
    the need to create these dim vectors manually.

    Parameters
    ----------
        data : ndarray
            The data that will be written to the EMD file. This is used to get the number of dims and their shape

        pixel_size : tuple, optional
            A tuple of pixel sizes. Must have same length as the number of dimensions of data.

    Returns
    -------
        dims: list
            A properly formatted tuple of dim vectors used as input
            to emd.emdFile.put_emdgroup()
    """
    if isinstance(pixel_size, np.ndarray):
        raise Exception('pixel_size must be a tuple')

    num = data.ndim

    if not pixel_size:
        pixel_size = (1,) * num

    if len(pixel_size) != data.ndim:
        raise ValueError('pixel_size and data dimensions must match')

    dims = []
    for ii in range(num):
        curDim = [np.linspace(0, data.shape[ii] - 1, data.shape[ii]) * pixel_size[ii],
                  'dim{}'.format(ii+1), 'unit{}'.format(ii+1)]
        dims.append(curDim)

    return dims


def emdReader(filename, dsetNum=0):
    """ A simple helper function to read in the data and metadata 
    in a structured format similar to the other ncempy readers.

    Note
    ----
        Note fully implemented yet. Work in progress.

    Parameters
    ----------
        filename : str or pathlib.Path
            The path to the file as a string.
        dsetNum : int
            The index of the data set to load.
            
    Returns
    -------
        : dict
            Data and metadata as a dictionary similar to other ncempy readers.

    Example
    -------
        Simply load all data and metadata from a data set in an EMD Velox file
            >> import ncempy.io as nio
            >> emd0 = nio.emd.emdReader('filename.emd', dsetNum = 0)

    """
    with fileEMD(filename, readonly=True) as emd0:
        d, dims = emd0.get_emdgroup(dsetNum, memmap=False)  # memmap must be false. File is closed
        out = {'data': d, 'filename': filename, 'pixelSize': []}

        for dim in dims:
            try:
                d = dim[0][1] - dim[0][0]
            except:
                d = 0
            out['pixelSize'].append(d)
        out['pixelUnit'] = [aa[2] for aa in dims]
        out['pixelName'] = [aa[1] for aa in dims]
        return out


def emdWriter(filename, data, pixel_size=None):
    """ Simple method to write data to a file formatted as an EMD v0.2. The only possible metadata to write is the pixel
    size for each dimension. Use the emd.fileEMD() class for more complex operations. The file must not already exist.

    Parameters
    ----------
    filename : str or pathlib.Path
        The path and file name to write the data to.
    data : ndarray
        The data as an ndarray to write to the file.
    pixel_size : tuple
        A tuple with the same length as the number of dims in data. len(pixel_size) == data.ndim

    """
    if isinstance(filename, str):
        filename = Path(filename)

    if pixel_size:
        try:
            assert len(pixel_size) == data.ndim
        except ValueError:
            raise ValueError('pixel_size length must match the number of dimensions of data.')

    if not filename.exists():
        with fileEMD(filename, readonly=False) as emd0:
            # Setup the dims for the EMD file. Pixel size is set to 1 by default for each dimension
            dims0 = defaultDims(data, pixel_size=pixel_size)
            # Write the data to he emd file.
            emd0.put_emdgroup('converted', data, dims0)
    else:
        raise FileExistsError



"""
A module to read MRC files in python and numpy.
Written according to MRC specification at http://bio3d.colorado.edu/imod/betaDoc/mrc_format.txt
Also works with FEI MRC files which include a special header block with experimental information.

Note
----
General users:
    Use the simplified mrc.mrcReader() function to load the data and meta
    data as a python dictionary.

Advanced users and developers:
    Access the file internals through the mrc.fileMRC() class.

"""

from pathlib import Path

import numpy as np


class fileMRC:
    """ Read in the data in MRC format and other useful information like metadata. Follows the specification
    published at http://bio3d.colorado.edu/imod/betaDoc/mrc_format.txt

    Attributes
    ----------
    fid : file
        The file handle to the opened MRC file.
    mrcType : int
        The internal MRC data type.
    dataType : np.dtype
        The numpy dtype corresponding to the mrcType.
    dataSize : np.ndarray
        The number of pixels along each dimension. Corresponds to the shape attribute of a np.ndarray
    gridSize : np.ndarray
        The size of the grid. Usually the same as dataSize
    volumeSize : np.ndarray
        The size of the volume along each direction in Angstroms.
    voxelSize : np.ndarray
        The size of the voxel along each direction in Angstroms.
    cellAngles : np.ndarray
        The angles of the cell. Ignored in most cases including ncempy.
    axisOrientations : np.ndarray
        Mapping the orientations of the data to real space directions X, Y, Z. Ignored by ncempy
    minMaxMean  : np.ndarray
        The minimum, maximum and mean value of the data to avoid computing this every time.
    extra : np.ndarray
        Extra mbinary metadata if it exists.
    FEIinfo : dict
        A dictionary of metadata used by FEI (Thermo Fischer) microsocpes for important metadata. This metadata
        overwrites the voxelsize attribute if it exists.
    dataOffset : int
        The integer offset in bytes to the start of the raw data.
    dataOut: dict
        Will hold the data and metadata to output to the user after getDataset() call.
    v : bool
        More output for debugging. False by default

    Examples
    --------
    Read in all data and metadata into memory.

    >>> import ncempy.io as nio
    >>> mrc0 = nio.mrc.mrcReader('file.mrc')

    Low level operations to get 1 slice of the 3D data

    >>> import ncempy.io as nio
    >>> with nio.mrc.fileMRC('file.mrc') as f1:
    >>>     single_slice = f1.getSlice(0)
    """

    def __init__(self, filename, verbose=False):
        """
        Parameters
        -----------
            filename: str or pathlib.Path
                String or pathlib.Path object pointing to the filesystem location of the file.
            verbose : bool
                If True, debug information is printed.

        """
        # check filename type
        if isinstance(filename, str):
            pass
        elif isinstance(filename, Path):
            filename = str(filename)
        else:
            raise TypeError('Filename is supposed to be a string or pathlib.Path')

        self.filename = filename

        # necessary declarations, if something fails
        self.fid = None
        self.mrcType = None
        self.dataType = None
        self.dataSize = None
        self.gridSize = None
        self.volumeSize = None
        self.voxelSize = None
        self.cellAngles = None
        self.axisOrientations = None
        self.minMaxMean = None
        self.extra = None
        self.FEIinfo = None
        self.dataOffset = None

        self.dataOut = {}  # will hold the data and metadata to output to the user after getDataset() call

        # Add a top level variable to indicate verbose output for debugging
        self.v = verbose

        # Open the file and quit if the file does not exist
        try:
            self.fid = open(self.filename, 'rb')
        except IOError as e:
            print("I/O error({0}): {1}".format(e.errno, e.strerror))

        # Store the original filename
        self.dataOut['filename'] = self.filename

        self.parseHeader()

    def __del__(self):
        """Close the file.

        """
        if not self.fid.closed:
            if self.v:
                print('Closing input file: {}'.format(self.filename))
            self.fid.close()
        return None

    def __enter__(self):
        """Implement python's with statement

        """
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Implement python's with statement.
        Close the file using __del__

        """
        self.__del__()
        return None

    def parseHeader(self):
        """Read the header information which includes data type, data size, data
        shape, and metadata.

        Note
        -----
            This header uses Fortran-style ordering. Numpy uses C-style ordering.
            The header is read in and then some attributes are reversed [::-1] at
            the end for output to the user to match C-ordering in numpy.

        TODO
        ----
            Implement special dtype to read the entire header at once.
            Read everything at once using special dtype. ~5x faster than multiple np.fromfile() reads:

            Untested but works in theory
            headerDtype = np.dtype([('head1','10int32'),('head2','6float32'),('axisOrientations','3int32'),
            ('minMaxMean','3int32'),('extra','32int32')])
            head = np.fromfile(self.fid,dtype=headerDtype,count=1)
        """

        # Always start at the beginning of the file.
        self.fid.seek(0)

        header_dtype = np.dtype(
            [('head1', '10int32'), ('head2', '6float32'), ('axisOrientations', '3int32'), ('minMaxMean', '3int32'),
             ('extra', '32int32')])
        header = np.fromfile(self.fid, dtype=header_dtype, count=1)

        # Read in the initial header values
        head1 = header['head1'][0]
        if self.v:
            print('header1 = {}'.format(head1))
        # Set the number of pixels for each dimension
        self.dataSize = head1[0:3]
        if self.v:
            print('dataSize (fortran ordering) = {}'.format(self.dataSize))

        # Set the data type and convert to numpy type
        self.mrcType = head1[3]
        self.dataType = self._getMRCType(self.mrcType)
        if self.v:
            print('dataType = {}'.format(self.dataType))

        # Get the grid size
        self.gridSize = head1[7:10]
        if self.v:
            print('mrc defined gridSize = {}'.format(self.gridSize))

        # Get the physical volume size (always in Angstroms) (starting at byte #11 in the file).
        head2 = header['head2'][0]
        self.volumeSize = head2[0:3]
        if self.v:
            print('mrc defined volumeSize = {}'.format(self.volumeSize))

        # calculate the voxel size based on volume and grid sizes
        # account for zero values which are often stored in volume and grid size values
        self.voxelSize = np.ones(3)
        for ii in range(0, 3):
            if self.volumeSize[ii] == 0:
                v = 1
            else:
                v = self.volumeSize[ii]
            if self.gridSize[ii] == 0:
                gs = 1
            else:
                gs = self.gridSize[ii]
            self.voxelSize[ii] = v / gs

        """if self.volumeSize.any() == 0 or self.gridSize.any() == 0:
            if self.v:
                print('Detected 0 volume or grid size. Setting voxel size to 1 (Ang).')
            self.voxelSize = np.ones(3)  # use 1 as a voxel size if its not set in the file
        else:
            self.voxelSize = (self.volumeSize / np.float32(self.gridSize))
            if self.v:
                print('voxelSize (Ang) = {}'.format(self.voxelSize))
        """

        # Pixel (cell) angles
        self.cellAngles = head2[3:6]
        if self.v:
            print('cellAngles = {}'.format(self.cellAngles))

        # Axis orientations. Tells which axes are X,Y,Z
        self.axisOrientations = header['axisOrientations'][0]
        if self.v:
            print('axisOrientations = {}'.format(self.axisOrientations))

        # Min, max,mean
        self.minMaxMean = header['minMaxMean'][0]

        # Extra information (for FEI MRC file, extra(1) is the size of the FEI information encoded with the file in
        # terms of 4 byte floats)
        self.extra = header['extra'][0]

        # Numpy uses C-style ordering. The header is written in Fortran-Style ordering.
        # Flip the order of everything useful
        if self.v:
            print(
                'Note: The MRC header is written in Fortran-Style ordering, but Numpy uses C-style ordering. '
                'This program will now reverse the order (using [::-1]) of useful metadata: (dataSize, gridSize, '
                'volumeSize, voxelSize, cellAngles, axisOrientations)')

        self.dataSize = self.dataSize[::-1]
        self.gridSize = self.gridSize[::-1]
        self.volumeSize = self.volumeSize[::-1]
        self.voxelSize = self.voxelSize[::-1]
        self.cellAngles = self.cellAngles[::-1]
        self.axisOrientations = self.axisOrientations[::-1]
        # self.Shape = self.Shape[::-1]

        # Move to the end of the normal header
        self.fid.seek(1024)

        # Read in the extended header if it exists (for FEI MRC files)
        if self.extra[1] != 0:
            pos1 = self.fid.tell()
            if self.v:
                print('Extra header found. Most likely and FEI-style MRC file.')
                print('Position before reading extra header: ' + str(pos1))
                print('Extra header size = ' + str(self.extra[1]))

            ''' Read the extra FEI header described as follows:
             1 a_tilt  first Alpha tilt (deg)
             2 b_tilt  first Beta tilt (deg)
             3 x_stage  Stage x position (Unit=m. But if value>1, unit=???m)
             4 y_stage  Stage y position (Unit=m. But if value>1, unit=???m)
             5 z_stage  Stage z position (Unit=m. But if value>1, unit=???m)
             6 x_shift  Image shift x (Unit=m. But if value>1, unit=???m)
             7 y_shift  Image shift y (Unit=m. But if value>1, unit=???m)
             8 defocus  starting Defocus Unit=m. But if value>1, unit=???m)
             9 exp_time Exposure time (s)
             10 mean_int Mean value of image
             11 tilt_axis   Tilt axis (deg)
             12 pixel_size  Pixel size of image (m)
             13 magnification   Magnification used
             14 voltage accelerating voltage
             15 ??
            '''
            FEIinfoValues = np.fromfile(self.fid, dtype=np.float32, count=15)
            self.FEIinfo = {'a_tilt': FEIinfoValues[0], 'b_tilt': FEIinfoValues[1], 'x_stage': FEIinfoValues[2],
                            'y_stage': FEIinfoValues[3], 'z_stage': FEIinfoValues[4], 'x_shift': FEIinfoValues[5],
                            'y_shift': FEIinfoValues[6], 'defocus': FEIinfoValues[7], 'exposure_time': FEIinfoValues[8],
                            'mean': FEIinfoValues[9], 'tilt_axis': FEIinfoValues[10], 'pixel_size': FEIinfoValues[11],
                            'magnification': FEIinfoValues[12], 'voltage': FEIinfoValues[13],
                            'unknown': FEIinfoValues[14]}

            self.voxelSize[0] = 1.  # set this to 1 but it should be the tilt angles. These can be non-uniform though.
            if self.FEIinfo['pixel_size'] != 0:
                self.voxelSize[1] = self.FEIinfo['pixel_size'] * 1e10  # convert meter to Angstroms, standard for MRCs
                self.voxelSize[2] = self.FEIinfo['pixel_size'] * 1e10

            if self.v:
                print('Extended header data')
                for aa, bb in self.FEIinfo.items():
                    print('{} = {}'.format(aa, bb))
        else:
            self.FEIinfo = {}

        self.dataOffset = 1024 + self.extra[1]  # offset of the data from the start of the file

        # Add relevant information (metadata) to the output dictionary
        self.dataOut = {'pixelSize': self.voxelSize, 'voxelSize': self.voxelSize,
                        'cellAngles': self.cellAngles, 'axisOrientations': self.axisOrientations,
                        'filename': self.filename}
        if self.extra[1] != 0:
            self.dataOut['FEIinfo'] = self.FEIinfo

        return 1

    def getDataset(self):
        """Read in the full data block and reshape to an ndarray
        with C-style ordering.

        """
        self.fid.seek(self.dataOffset, 0)  # move to the start of the data from the start of the file
        try:
            num0 = int(np.prod(self.dataSize, dtype=np.uint64))
            data1 = np.fromfile(self.fid, dtype=self.dataType, count=num0)
            self.dataOut['data'] = data1.reshape(self.dataSize)
        except MemoryError:
            print("Not enough memory to read in the full data set. Use getMemmap")
        return self.dataOut

    def getSlice(self, num):
        """Read in a slice of an MRC file. Useful for parsing through a large file without reading
        the entire data set into memory.

        Parameters
        ----------
            num : int
                Get the requested image.

        Returns
        -------
            out : ndarray
                A 2D slice or a 3D set of slices along the first index

        Raises
        ------
            IndexError
                If num > the number of slices.

        """
        # Check num is within the data array size bounds
        if num > (self.dataSize[0] - 1):
            raise IndexError('Index {} is out of bounds for array with size {}'.format(num, self.dataSize[0]))

        self.fid.seek(self.dataOffset, 0)  # move to the start of the data by skipping the header
        imSize = self.dataSize[1] * self.dataSize[2]  # size of each image in pixels
        byteSize = int(np.dtype(self.dataType).itemsize * imSize)  # make sure this is an int32
        self.fid.seek(num * byteSize, 1)  # skip to the slice requested from the start of the data

        data1 = np.fromfile(self.fid, dtype=self.dataType, count=imSize)  # read in the requested image
        data1 = data1.reshape((self.dataSize[1], self.dataSize[2]))  # reshape the image

        return data1

    def getMemmap(self):
        """Return a numpy memmap object (read-only) for the dataset. This is very useful
        for very large datasets to avoid loading the entire data set into memory. No meta data is
        returned.

        Returns
        --------
        : numpy.core.memmap
            A read-only numpy memmap object with access to the data on disk.
        """
        mm = np.memmap(self.fid, dtype=self.dataType, mode='r', offset=self.dataOffset,
                       shape=tuple(self.dataSize))

        return mm

    def _applyAxisOrientations(self, arrayIn):
        """ This is untested and unused.

        """
        return [arrayIn[x - 1] for x in self.axisOrientations]

    def _getMRCType(self, dataType):
        """Return the correct data type according to the official MRC type list:
        0 image : signed 8-bit bytes range -128 to 127
        1 image : 16-bit
        2 image : 32-bit reals
        3 transform : complex 16-bit integers
        4 transform : complex 32-bit reals
        6 image : unsigned 16-bit range 0 to 65535

        Parameters
        ----------
            dataType: int
                The data type value encoded in an MRC header

        Returns
        --------
            out: numpy dtype
                The corresponding numpy data type.

        """
        if dataType == 0:
            Type = np.int8
        elif dataType == 1:
            Type = np.int16
        elif dataType == 2:
            Type = np.float32
        elif dataType == 6:
            Type = np.uint16
        else:
            print("Unsupported data type" + str(dataType))  # complex data types are currently unsupported
            Type = None
        return Type


def mrcReader(file_name, verbose=False):
    """A simple function to read open a MRC, parse the header, and read the full
    data set.

    Parameters
    ----------
        file_name: str
            The name of the file to load
        verbose : bool, optional
            Enable printing debug messages as the header is parsed.

    Returns
    -------
        out: dict
            A dictionary containing the data and interesting metadata. The data is attached to the 'data' key.

    Example
    -------
        Simply read in all data from disk into memory. This assumes the dataset is 3 dimensional:

            >>> from ncempy.io import mrc
            >>> import matplotlib.pyplot as plt
            >>> mrc1 = mrc.mrcReader('filename.mrc')
            >>> plt.imshow(mrc1['data'][0, :, :]) #show the first image in the data set
    """
    with fileMRC(file_name, verbose) as f1:  # open the file and init the class
        im1 = f1.getDataset()  # read in the dataset

    im1['pixelUnit'] = 'A'

    return im1  # return the data and metadata as a dictionary


def mrc2raw(file_name):
    """ Convert an MRC type file to raw binary. The entire data is read into memory and then written to a new
    raw file in the same location with the data type and shape (C-ordering) written into the filename. No other
    meta data is retained.

    Parameters
    ----------
    file_name : str or pathlib.Path
        The file name to convert.

    """
    if isinstance(file_name, str):
        file_name = Path(file_name)

    tomo = mrcReader(file_name)

    # Add metadata to file name
    out_name = file_name.stem + '_{}_{}.raw'.format(tomo['data'].dtype, tomo['data'].shape)

    outPath = file_name.with_name(out_name.replace(' ', ''))
    print(outPath)

    with open(outPath, 'wb') as f0:
        f0.write(tomo['data'])  # write out as C ordered data


def mrc2emd(file_name):
    """Write an MRC file as an HDF5 file in EMD format with same file name and .emd ending.
    Header information is retained as attributes.

    Parameters
    ----------
        file_name: str
            The name of the file to convert from MRC to EMD format.

    Returns
    -------
        out: int
            1 if successful.

    TODO
    -----
        Update this to use ncempy.emd class
    """
    import h5py

    # Read in the MRC data and reshape to C-style ordering
    tomo = mrcReader(file_name)

    # create the HDF5 file
    with h5py.File(file_name.rsplit('.mrc', 1)[0] + '.emd', 'w') as f1:  # w- will error if the file exists

        # Create the axis vectors in nanometers. Standard MRC pixel size is in Angstroms
        xFull = np.linspace(0, tomo['voxelSize'][0] * tomo['data'].shape[0] - 1, tomo['data'].shape[0])
        yFull = np.linspace(0, tomo['voxelSize'][1] * tomo['data'].shape[1] - 1, tomo['data'].shape[1])
        zFull = np.linspace(0, tomo['voxelSize'][2] * tomo['data'].shape[2] - 1, tomo['data'].shape[2])

        # Root data group
        dataTop = f1.create_group('data')

        # Create tilt series group
        tiltseriesGroup = dataTop.create_group('data')
        tiltseriesGroup.attrs['emd_group_type'] = np.int8(1)

        # Save the data to the EMD file and reshape it to a C-style array
        try:
            _ = tiltseriesGroup.create_dataset('data', data=tomo['data'], compression='gzip', shuffle=True)
        except MemoryError:
            raise MemoryError("Not enough memory to write out data to EMD file")

        dim1 = tiltseriesGroup.create_dataset('dim1', data=xFull)
        dim1.attrs['name'] = np.string_('x')
        dim1.attrs['units'] = np.string_('')
        dim2 = tiltseriesGroup.create_dataset('dim2', data=yFull)
        dim2.attrs['name'] = np.string_('y')
        dim2.attrs['units'] = np.string_('')
        dim3 = tiltseriesGroup.create_dataset('dim3', data=zFull)
        dim3.attrs['name'] = np.string_('z')
        dim3.attrs['units'] = np.string_('')

        # Create the other groups
        scopeGroup = f1.create_group('Microscope')
        scopeGroup.attrs['voxel sizes'] = tomo['voxelSize']
        userGroup = f1.create_group('User')
        commentGroup = f1.create_group('Comments')


def mrcWriter(filename, data, pixelSize, forceWrite=False):
    """Write out a MRC type file according to the specification at http://bio3d.colorado.edu/imod/doc/mrc_format.txt

    Parameters
    ----------
        filename : str or pathlib.Path
            The name or Path of the file to write out to.
        data : ndarray
            The array data to write to disk.
        pixelSize : tuple
            The size of the pixel along each direction (in Angstroms) as a 3 element vector (sizeZ,sizeY,sizeX).
        forceWrite : bool
            This will write the data as a C-contiguous array. It is not suggested to use this option and it will
            be removed in future versions.

    """

    with open(filename, 'wb') as fid:

        if len(data.shape) > 3:
            print("Too many dimensions")
            return

        if not data.flags['C_CONTIGUOUS']:
            print("Error: Array must be C-style ordering: [numImages,Y,X]. "
                  "Use numpy.tranpspose and np.ascontiguousarray to change data ordering in memory")
            print('Exiting')
            return

        # initialize the header with 256 zeros with size 4 bytes
        header = np.zeros(256, dtype=np.int32)
        fid.write(header)
        fid.seek(0, 0)  # return to the beginning of the file

        # Initialize the int32 part of the header
        header1 = np.zeros(10, dtype=np.int32)

        # Write the number of columns, rows and sections (images)
        # header1[0:3] = np.int32(dims) #data size in pixels
        header1[0] = np.int32(data.shape[2])  # num columns, the last index in C-style ordering
        header1[1] = np.int32(data.shape[1])  # num rows
        header1[2] = np.int32(data.shape[0])  # num sections (images)

        if data.dtype == np.float32:
            header1[3] = np.int32(2)
        elif data.dtype == np.uint16:
            header1[3] = np.int32(6)
        elif data.dtype == np.int16:
            header1[3] = np.int32(1)
        elif data.dtype == np.int8:
            header1[3] = np.int32(0)
        else:
            print("Data type {} is unsupported. Only int8, int16, uint16, and float32 are supported".format(data.dtype))
            return

        # Starting point of sub image (not used in IMOD)
        header1[4:7] = np.zeros(3, dtype=np.int32)

        # Grid size in X,Y,Z
        # header1[7:10] = np.int32(dims); #data size in pixels
        header1[7] = np.int32(data.shape[2])  # mx
        header1[8] = np.int32(data.shape[1])  # my
        header1[9] = np.int32(data.shape[0])  # mz

        # Write out the first part of the header information
        fid.write(header1)

        # Cell dimensions (in Angstroms)
        # pixel spacing = xlen/mx, ylen/my, zlen/mz
        fid.write(np.float32(pixelSize[2] * data.shape[2]))  # xlen
        fid.write(np.float32(pixelSize[1] * data.shape[1]))  # ylen
        fid.write(np.float32(pixelSize[0] * data.shape[0]))  # zlen

        # Cell angles (in degrees)
        fid.write(np.float32([90.0, 90.0, 90.0]))

        # Description of array directions with respect to: Columns, Rows, Images
        fid.write(np.int32([1, 2, 3]))

        # Minimum and maximum density
        fid.write(np.float32(np.min(data)))
        fid.write(np.float32(np.max(data)))
        fid.write(np.float32(np.mean(data)))

        # Needed to indicate that the data is little endian for NEW-STYLE MRC image2000 HEADER - IMOD 2.6.20 and above
        fid.seek(212, 0)
        fid.write(np.int8([68, 65, 0, 0]))  # use [17,17,0,0] for big endian

        # Write out the data
        fid.seek(1024)
        if forceWrite:
            fid.write(np.ascontiguousarray(data))  # Change to C ordering array for writing to disk
        else:
            fid.write(data)  # must be C-contiguous

        # Close the file


def writeHeader(filename, shape, dtype, pixelSize):
    """Write out a MRC type file header according to the specification at
    http://bio3d.colorado.edu/imod/doc/mrc_format.txt. This is useful for initializing an MRC file and then writing to
    it manually or see appendData() function below.

    Parameters
    ----------
    filename : str
        The name of the EMD file
    shape : tuple
        The shape of the data to write
    dtype : numpy.dtype
        The dtype to write out the data as. Only some numpy dtypes are supported byt his format. It is suggested
        to use np.float32 in most cases for maximum compatibility.
    pixelSize : tuple
        The size of the pixel along each direction (in Angstroms) as a 3 element vector (sizeX,sizeY,sizeZ).
        sizeZ could be the angular step for a tilt series

    """

    with open(filename, 'wb') as fid:
        if len(shape) > 3:
            print("Too many dimensions")
            return 0

        # Initialize the header with 256 zeros with size 4 bytes
        header = np.zeros(256, dtype=np.int32)
        fid.write(header)
        fid.seek(0, 0)  # return to the beginning of the file

        # Initialize the int32 part of the header with zeros
        header1 = np.zeros(10, dtype=np.int32)

        # Write the number of columns, rows and sections (images)
        header1[0] = np.int32(shape[2])  # num columns, the last index in C-style ordering
        header1[1] = np.int32(shape[1])  # num rows
        header1[2] = np.int32(shape[0])  # num sections (images)

        if dtype == np.float32:
            header1[3] = np.int32(2)
        elif dtype == np.uint16:
            header1[3] = np.int32(6)
        elif dtype == np.int16:
            header1[3] = np.int32(1)
        elif dtype == np.int8:
            header1[3] = np.int32(0)
        else:
            print("Data type " + str(dtype) + " is unsupported. Only int8, int16, uint16, and float32 are supported")
            return 0

        # Starting point of sub image (not used in IMOD)
        header1[4:7] = np.zeros(3, dtype=np.int32)

        # Grid size in X,Y,Z
        header1[7] = np.int32(shape[2])  # mx
        header1[8] = np.int32(shape[1])  # my
        header1[9] = np.int32(shape[0])  # mz

        # Write out the first part of the header information
        fid.write(header1)

        # Cell dimensions (in Angstroms)
        # pixel spacing = xlen/mx, ylen/my, zlen/mz
        fid.write(np.float32(pixelSize[2] * shape[2]))  # xlen
        fid.write(np.float32(pixelSize[1] * shape[1]))  # ylen
        fid.write(np.float32(pixelSize[0] * shape[0]))  # zlen

        # Cell angles (in degrees)
        fid.write(np.float32([90.0, 90.0, 90.0]))

        # Description of array directions with respect to: Columns, Rows, Images
        fid.write(np.int32([1, 2, 3]))

        # Minimum and maximum density
        # fid.write(np.float32(np.min(data)))
        # fid.write(np.float32(np.max(data)))
        # fid.write(np.float32(np.mean(data)))
        fid.write(np.zeros(3, dtype=np.float32))

        # Needed to indicate that the data is little endian for NEW-STYLE MRC image2000 HEADER - IMOD 2.6.20 and above
        fid.seek(212, 0)
        fid.write(np.int8([68, 65, 0, 0]))  # use [17,17,0,0] for big endian


def appendData(filename, data):
    """Append a binary set of data to the end of a MRC file. This should only be used in conjunction with
    writeHeader() above.

    Parameters
    ----------
    filename : str
        Name of the MRC file with pre-initiated header and some data already written.
    data : ndarray
        Data to append to the file.

    """
    with open(filename, 'ab') as fid:
        # Write out the data
        fid.seek(0, 2)  # seek to the end of the file
        fid.write(data)  # Change to C ordering array for writing to disk


def emd2mrc(filename, dsetPath):
    """Convert EMD data set into MRC data set. The final data type is float32 for convenience.

    Parameters
    ----------
    filename : str
        The name of the EMD file
    dsetPath : str
        The HDF5 path to the top group holding the data. ex. '/data/raw/'
    """
    import h5py
    with h5py.File(filename, 'r') as f1:
        # Get the pixel sizes and convert to Ang
        dimsPath = dsetPath + '/dim'
        print('Warning: Assuming EMD dim vectors are in nanometer')
        print('dim vector names/units are:')
        for ii in range(1, 4):
            print('name, units = {}, {}'.format(f1[dimsPath + str(ii)].attrs['name'],
                                                f1[dimsPath + str(ii)].attrs['units']))
        pixelSizeX = (f1[dsetPath + '/dim2'][1] - f1[dsetPath + '/dim2'][0]) * 10  # change nanometers to Ang
        pixelSizeY = (f1[dsetPath + '/dim3'][1] - f1[dsetPath + '/dim3'][0]) * 10  # change nanometers to Ang

        filenameOut = filename.split('.emd')[
                          0] + '.mrc'  # use the first part of the file as the prefix removing the .emd on the end

        print('Warning: Converting to float32 before writing to disk')
        mrcWriter(filenameOut, np.float32(f1[dsetPath + '/data']),
                  (1, pixelSizeY, pixelSizeX))

        print('Finished writing to: {}'.format(filenameOut))



"""
A set of visualization functions based on matplotlib.pyplot which are generally
useful for S/TEM data visualization.

"""

import numpy as np
import matplotlib.pyplot as plt

import ncempy.algo
from ncempy.algo.distortion import rad_dis


def imsd(im, vmin=-2, vmax=2, **kwargs):
    """Show an array as an image with intensities compared to the standard deviation of the data. Other
    keyword args are passed to pyplot.imshow(). vmin and vmax are by default set to -2 and 2 respectively
    which are usually good values to set for S/TEM data.

    Parameters
    ----------
    im : np.ndarray
        The image to show.

    vmin, vmax : float, default = -2, 2
        The vmin and vmax values to pass to imshow.

    Returns
    -------
    : matplotlib.pyplot.Figure
        The handle to the created figure
    """
    fg, ax = plt.subplots(1, 1)
    im2 = im - im.mean()
    im3 = im2 / np.std(im2)
    imax = ax.imshow(im3, vmin=vmin, vmax=vmax, **kwargs)
    return imax


def im_calibrated(im, d):
    """ Plot an image calibrated using the pixel size d. The centers of the pixels will be the
    the center of each measurement. So, if you plot positions in real coordinates the points
    will be plotted in the center of the pixel.

    Parameters
    ---------
    im : np.ndarray
        The image to show using imshow
    d : float
        The pixel size in both directions. The pixel size must be isotropic.

    Returns
    -------
    : pyplot.figure
        The figure containing the plot
    """
    # The default extent is calculated like this:
    ext = [-0.5, im.shape[1] - 0.5, im.shape[0] - 0.5, -0.5]
    # Calibrate the extent
    ext = [ii * d for ii in ext]

    fg, ax = plt.subplots(1, 1)
    ax.imshow(im, extent=ext)
    return fg

def imfft(im, d=1.0, ax=None):
    """ Show a 2D FFT as a diffractogram with log scaling applied and zero frequency
    fftshifted tp the center. A new figure is created or an axis can be specified.

    The diffracotgram is calculated from the original intensities (I) as

    .. math::
        1 + 0.001 * I ^2

    Parameters
    ----------
    im: np.ndarray
        The 2D fft of the diffraction pattern to display as a diffractogram
    d: float, optional, default = 1.0
        The real space pixel size of the image used to get the FFT
    ax: pyplot axis, optional
        An axis to plot into.
    Returns
    -------
    : matplotlib.image.AxesImage
        The AxesImage that contains the image displayed.

    Example
    -------
    This example shows how to display a 2D ndarray (image) as a
    diffractogram. The image has a real space pixel size of 0.1 nanometer.

    >>> imageFFT = np.fft.fft2(im)
    >>> ncempy.viz.imfft(imageFFT, d = 0.1)

    """

    fftFreq0 = np.fft.fftshift(np.fft.fftfreq(im.shape[0], d))
    fftFreq1 = np.fft.fftshift(np.fft.fftfreq(im.shape[1], d))
    if ax is None:
        fg, ax = plt.subplots(1, 1)
    imax = ax.imshow(np.fft.fftshift(np.log(1 + .001 * np.abs(im) ** 2)),
                     extent=(fftFreq0[0], fftFreq0[-1], fftFreq1[-1], fftFreq1[0]))
    return imax


def imrfft(im, d=1.0, ax=None):
    """Show a 2D rFFT (real FFT) as a diffractogram with log scaling applied
    and fftshift-ed along axis 0. See imfft for full details.

    Parameters
    ----------
    im : ndarray
        The 2D fft of the diffraction pattern to display as a diffractogram
    d : float, optional, default = 1.0
        The real space pixel size of the image used to get the FFT
    ax : pyplot axis, optional
        An axis to plot into.
    Returns
    -------
    : matplotlib.image.AxesImage
        The AxesImage that contains the image displayed.
    """

    fftFreq1 = np.fft.fftshift(np.fft.fftfreq(im.shape[1], d))
    fftFreq0 = np.fft.rfftfreq(im.shape[0], d)
    if ax is None:
        fg, ax = plt.subplots(1, 1)
    axim = ax.imshow(np.fft.fftshift(np.log(1 + .001 * np.abs(im) ** 2), axes=0),
              extent=(fftFreq0[0], fftFreq0[-1], fftFreq1[-1], fftFreq1[0]))

    return axim


def im_and_fft(im, d=1.0, fft=None):
    """ Show the image and its fft side by side. Uses imfft to show the fft.

    Parameters
    ----------
    im : np.ndarray
        The image to show in both real and FFT space
    d : float
        The pixel spacing
    fft : np.ndarray, optional
        The FFT to display. If not provided then np.fft.fft2 is used.

    Returns
    -------
    : plt.figure
        The matplotlib.pyplot figure
    """
    fg, ax = plt.subplots(1, 2)
    ax[0].imshow(im)

    if not fft:
        fft = np.fft.fft2(im)
    imfft(fft, d=d, ax=ax[1])


class stack_view:
    """
    Class to allow a volume to be scrubbed through with a matplotlib slider widget.
    The first axis of the volume is the slicing axis. Other keyword args are passed
    directly to imshow upon the figure creation.

    Parameters
    ----------
    stack : numpy.ndarray, 3D stack
        The stack of to show as images

    **kwargs :
        Passed directly to pyplot.imshow()

    """

    def __init__(self, stack, **kwargs):
        from matplotlib.widgets import Slider

        if stack.ndim != 3:
            raise Exception('Must be three-dimensional stack of images.')

        self.fig, self.ax = plt.subplots()
        plt.subplots_adjust(left=0.25, bottom=0.25)

        self._st = stack  # internal reference to the stack

        # Initialize the imshow axis
        self.im0 = int(self._st.shape[0] / 2)  # initial slice to show
        self.axI = self.ax.imshow(stack[self.im0, :, :], **kwargs)

        # Setup the slider
        ax_color = 'lightgoldenrodyellow'
        self.axSlider = plt.axes([0.25, 0.1, 0.65, 0.03], facecolor=ax_color)
        self.sl = Slider(self.axSlider, 'Slice', 0, self._st.shape[0] - 1, valinit=self.im0, valfmt='%1.f')
        self.sl.on_changed(self._update)

        plt.show()

    def _update(self, val):
        num = self.sl.val
        self.axI.set_data(self._st[int(round(num)), :, :])
        self.fig.canvas.draw_idle()


def plot_ringpolar(points, dims, show=False):
    """Plot points in polar coordinate system.

    Parameters
    ----------
    points : np.ndarray
        Positions in polar coords.
    dims : tuple
        Dimension information to plot labels.
    show : bool
        Set to directly show plot in interactive mode.

    Returns
    -------
    : numpy.ndarray
        Image of the plot.

    """

    try:
        # try to convert input to np.ndarray with 2 columns (necessary if only one entry provided)
        points = np.reshape(np.array(points), (-1, 2))
        # check if enough dims available
        assert (len(dims) >= 2)
        assert (len(dims[0]) == 3)
    except:
        raise TypeError('Something wrong with the input!')

    fig = plt.figure()
    ax = fig.add_subplot(111)

    # mean value as line
    ax.axhline(np.mean(points[:, 0]), ls='--', c='k')

    # points
    ax.plot(points[:, 1], points[:, 0], 'rx')

    # labels
    ax.set_xlabel('theta /[rad]')
    ax.set_xlim((-np.pi, np.pi))
    ax.set_ylabel('r /{}'.format(dims[0][2]))

    if show:
        plt.show(block=False)

    # render to array
    fig.canvas.draw()
    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))

    return plot


def plot_distpolar(points, dims, dists, ns, show=False):
    """Plot the results of distortion fitting in polar coordinates.

    Parameters
    ----------
    points : np.ndarray
        Points in polar coords.

    dims : tuple
        Dimensions, necessary to have unit information.

    dists : np.ndarray
        Results of dist fitting, length according to ns.

    ns : list
        List of used orders.

    show : bool
        Set to directly show the plot in interactive mode.

    Returns
    -------
    : np.ndarray
        Image of the plot.

    """

    try:
        # check points
        assert (isinstance(points, np.ndarray))
        assert (points.shape[1] == 2)

        # check if enough dims available
        assert (len(dims) >= 2)
        assert (len(dims[0]) == 3)

        # check orders
        assert (len(ns) >= 1)

        # check dists
        assert (dists.shape[0] == len(ns) * 2 + 1)
    except:
        raise TypeError('Something wrong with the input!')

    fig = plt.figure()
    ax = fig.add_subplot(111)

    # stuff from the single orders
    ax.axhline(dists[0], ls='--', c='k')
    xpl_ell = np.linspace(-np.pi, np.pi, 100)
    for i in range(len(ns)):
        plt.plot(xpl_ell, dists[0] * rad_dis(xpl_ell, dists[i * 2 + 1], dists[i * 2 + 2], ns[i]), 'm--')

    # points before
    ax.plot(points[:, 1], points[:, 0], 'rx')

    # sum of all distorts
    sum_dists = np.ones(xpl_ell.shape) * dists[0]
    for i in range(len(ns)):
        sum_dists *= rad_dis(xpl_ell, dists[i * 2 + 1], dists[i * 2 + 2], ns[i])
    plt.plot(xpl_ell, sum_dists, 'b-')

    # points after
    points_corr = np.copy(points)
    for i in range(len(ns)):
        points_corr[:, 0] /= rad_dis(points[:, 1], dists[i * 2 + 1], dists[i * 2 + 2], ns[i])
    plt.plot(points_corr[:, 1], points_corr[:, 0], 'gx')

    # labels
    ax.set_xlabel('theta /[rad]')
    ax.set_xlim((-np.pi, np.pi))
    ax.set_ylabel('r /{}'.format(dims[0][2]))

    if show:
        plt.show(block=False)

    # render to array
    fig.canvas.draw()
    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))

    return plot


def plot_points(img, points, vminmax=(0, 1), dims=None, invert=False, show=False):
    """Plot the detected points on the input image for checking.

    Parameters
    ----------
    img : np.ndarray
        Image.
    points : np.ndarray
        Array containing the points.
    vminmax : tuple
        Tuple of two values for relative lower and upper cut off to display image.
    dims : tuple
        Tuple of dims to plot in dimensions.
    invert : bool
        Set to invert the image.
    show : bool
        Set to directly show the plot interactively.

    Returns
    -------
    : np.ndarray
        Image of the plot.

    """

    try:
        assert (isinstance(img, np.ndarray))

        assert (isinstance(points, np.ndarray))
        assert (points.shape[1] == 2)
        assert (len(points.shape) == 2)
    except:
        raise TypeError('Something wrong with the input!')

    fig = plt.figure()
    ax = fig.add_subplot(111)

    if invert:
        cmap = "Greys"
    else:
        cmap = "gray"

    if dims:
        ax.imshow(img, cmap=cmap, vmin=np.min(img) + vminmax[0] * (np.max(img) - np.min(img)),
                  vmax=np.min(img) + vminmax[1] * (np.max(img) - np.min(img)),
                  extent=(np.min(dims[0][0]), np.max(dims[0][0]), np.max(dims[1][0]), np.min(dims[1][0])))
        ax.set_xlabel('{} {}'.format(dims[0][1], dims[0][2]))
        ax.set_ylabel('{} {}'.format(dims[1][1], dims[1][2]))
        ax.set_xlim((np.min(dims[0][0]), np.max(dims[0][0])))
        ax.set_ylim((np.max(dims[1][0]), np.min(dims[1][0])))
    else:
        ax.imshow(img, cmap=cmap, vmin=np.min(img) + vminmax[0] * (np.max(img) - np.min(img)),
                  vmax=np.min(img) + vminmax[1] * (np.max(img) - np.min(img)))
        ax.set_xlim((0, img.shape[1] - 1))
        ax.set_ylim((img.shape[0] - 1, 0))

    ax.scatter(points[:, 1], points[:, 0], color='r', marker='o', facecolors='none')

    if show:
        try:
            plt.show(block=False)
        except:
            pass

    fig.canvas.draw()

    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))

    return plot


def plot_radialprofile(r, intens, dims, show=False):
    """Plot radial profile.

    Parameters
    ----------
    r : np.ndarray
        r-axis of radial profile.
    intens : np.ndarray
        Intensity-axis of radial profile.
    dims : tuple
        Dimensions of original image to read out units.
    show : bool
        Set to directly show plot interactively.

    Returns
    -------
    : np.ndarray
        Image of the plot.
    """

    try:
        # check data
        assert (isinstance(r, np.ndarray))
        assert (isinstance(intens, np.ndarray))
        assert (np.array_equal(r.shape, intens.shape))

        # check if dims available
        assert (len(dims) >= 1)
        assert (len(dims[0]) == 3)

    except:
        raise TypeError('Something wrong with the input!')

    fig = plt.figure()
    ax = fig.add_subplot(111)

    ax.plot(r, intens, 'r-')

    # labels
    ax.set_xlabel('r /{}'.format(dims[0][2]))
    ax.set_ylabel('I /[a.u.]')

    if show:
        plt.show(block=False)

    # render to array
    fig.canvas.draw()
    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))

    return plot


def plot_fit(r, intens, dims, funcs, param, show=False):
    """Plot the fit results to the radial profile.

    Parameters
    ----------
    r : np.ndarray
        r-axis of radial profile.
    intens : np.ndarray
        Intensity-axis of radial profile.
    dims : tuple
        Dimensions of original image to read out units.
    funcs : tuple
        List of functions.
    param : np.ndarray
        Parameters for functions in funcs.
    show : bool
        Set to directly show plot interactively.

    Returns
    -------
    : np.ndarray
        Image of the plot.

    """

    try:
        # check data
        assert (isinstance(r, np.ndarray))
        assert (isinstance(intens, np.ndarray))
        assert (np.array_equal(r.shape, intens.shape))

        # check if dims available
        assert (len(dims) >= 1)
        assert (len(dims[0]) == 3)

        # funcs and params
        assert (len(funcs) >= 1)
        for i in range(len(funcs)):
            assert (funcs[i] in ncempy.algo.math.lkp_funcs)

        param = np.array(param)
        param = np.reshape(param, sum(map(lambda x: ncempy.algo.math.lkp_funcs[x][1], funcs)))

    except:
        raise TypeError('Something wrong with the input!')

    fig = plt.figure()
    ax = fig.add_subplot(111)

    # plot radial profile
    ax.plot(r, intens, 'r-')

    # plot single
    n = 0
    for i in range(len(funcs)):
        ax.plot(r, ncempy.algo.math.lkp_funcs[funcs[i]][0](r, param[n:n + ncempy.algo.math.lkp_funcs[funcs[i]][1]]),
                'g-')
        n += ncempy.algo.math.lkp_funcs[funcs[i]][1]
    # sum of functions
    sum_funcs = ncempy.algo.math.sum_functions(r, funcs, param)
    ax.plot(r, sum_funcs, 'b-')

    # labels
    ax.set_xlabel('r /{}'.format(dims[0][2]))
    ax.set_ylabel('I /[a.u.]')

    if show:
        plt.show(block=False)

    # render to array
    fig.canvas.draw()
    plot = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    plot = plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))

    return plot



import numpy as np
from scipy import ndimage


def image_cross_corr(image, reference, real_filter=1, k_filter=1):
    """ Calculate image cross-correlation. See imageCrossCorRealShift and other
    similar functions to calculate the shift and apply the shift.

    Parameters
    ----------
        image : ndarray
            The source image to align to. Should be even sized.

        reference : ndarray
            The reference image to align to the source image. Should be even sized.

        real_filter : ndarray, optional, default 1
            Filter to apply to each image in real space before FFT.

        k_filter : ndarray, optional default 1
            Filter to apply to each image in FFT space

    Returns
    -------
        : ndarray
            Cross correlation of image and reference.

    """

    if type(image) is not np.ndarray or type(reference) is not np.ndarray:
        raise TypeError("Must use ndarrays")

    if np.iscomplexobj(image) or np.iscomplexobj(reference):
        raise TypeError("Images mst be real")

    # Calculate FFT of images
    image_f = np.fft.fft2((image - np.mean(image)) * real_filter)
    reference_f = np.fft.fft2((reference - np.mean(reference)) * real_filter)

    # Calculate cross-correlation
    xcor = abs(np.fft.fftshift(np.fft.ifft2(np.conj(image_f) * reference_f * k_filter)))

    return xcor


def image_correlate(image, reference, real_filter=1, k_filter=1, shift_func='shift', verbose=False):
    """ Align image to reference by cross-correlation. Outputs shifts and shifted images.
    Uses the real FFT for ~2x speed improvement. The k_filter must have
    a shape that matches the np.fft.rfft2() of image and reference.
    Uses scipy.ndimage.shift() or np.roll to move the image. Use 'roll' to avoid losing
    data off the edge for multiple shifting operations. Use shift to avoid wrap around problems and when there
    is only one shifting operation.

    Note
    ----
        image, reference and real_filter must all have the same shape (N, M).
        k_filter must have a shape that matches the np.fft.rfft2() of
        the other inputs: (N, M/2+1)

    Parameters
    ----------
        image : ndarray
            A image as a 2D ndarray.

        reference : ndarray
            The reference image to align to.

        real_filter : ndarray, optional, default = 1
            A real space filter applied to image and reference before
            calculating the shift.

        k_filter : ndarray, optional, default = 1
            A Fourier space filter applied to the fourier transform of
            image and reference before calculating the cross-correlation.

        shift_func : str, default is 'shift'
            The function to use to shift the images. 'roll' uses np.roll and 'shift' uses ndimage.shift.

        verbose : bool
            Plots the cross-correlation using matplotlib for debugging purposes.

    Returns
    -------
        : tuple, (ndarray, tuple)
            A tuple containing the shifted image and the shifts applied.
    """
    output = None

    if shift_func is not 'shift' and shift_func is not 'roll':
        raise KeyError('Shift function has to be either shift or roll')

    image_f = np.fft.rfft2((image - np.mean(image)) * real_filter)
    reference_f = np.fft.rfft2((reference - np.mean(reference)) * real_filter)

    xcor = abs(np.fft.irfft2(np.conj(image_f) * reference_f * k_filter))
    if verbose:
        import matplotlib.pyplot as plt
        plt.imshow(np.fft.fftshift(xcor))
        plt.title('imageCrossCorRealShift xcor')
    shifts = np.unravel_index(np.fft.fftshift(xcor).argmax(), xcor.shape)
    shifts = (shifts[0] - xcor.shape[0] / 2, shifts[1] - xcor.shape[1] / 2)
    shifts = [int(i) for i in shifts]  # convert to integers

    if shift_func == 'shift':
        # shift image using ndimage.shift
        output = ndimage.interpolation.shift(image, shifts, order=0)
    elif shift_func == 'roll':
        # shift image using roll to be reversible
        output = np.roll(image, shifts[0], axis=0)
        output = np.roll(output, shifts[1], axis=1)

    return output, shifts


def image_phase_correlate(image, reference, real_filter=1, k_filter=1, shift_func='shift', verbose=False):
    """ Align image to reference by phase-correlation. Outputs shifted images and shift.
    Uses np.fft.rfft2 for ~2x speed improvement.
    Uses scipy.ndimage.shift() to shift the image and remove border pixels.

    NOT WORKING OR TESTED YET

    Note
    ----
        image, reference and real_filter must all have the same shape (N, M).
        k_filter must have a shape that matches the np.fft.rfft2() of
        the other inputs: (N, M/2+1)

    Parameters
    ----------
        image : ndarray
            A image as a 2D ndarray.

        reference : ndarray
            The reference image to align to.

        real_filter : ndarray, optional, default = 1
            A real space filter applied to image and reference before
            calculating the shift.

        k_filter : ndarray, optional, default = 1
            A Fourier space filter applied to the fourier transform of
            image and reference before calculating the cross-correlation.

        shift_func : str, default is 'shift'
            The function to use to shift the images. 'roll' uses np.roll and 'shift' uses ndimage.shift.


        verbose : bool
            Plots the cross-correlation using matplotlib for debugging purposes.

    Returns
    ------
        : tuple, (ndarray, tuple)
            A tuple containing the shifted image and the shifts applied.

    """
    output = None

    image_f = np.fft.rfft2((image - np.mean(image)) * real_filter)
    reference_f = np.fft.rfft2((reference - np.mean(reference)) * real_filter)

    xcor = abs(np.fft.irfft2(np.conj(image_f) * reference_f * k_filter))
    pcor = xcor / (np.abs(xcor) + 0.001)
    if verbose:
        import matplotlib.pyplot as plt
        plt.imshow(np.fft.fftshift(pcor))
        plt.title('imageCrossPhaseRealShift pcor')
    shifts = np.unravel_index(np.fft.fftshift(pcor).argmax(), pcor.shape)
    shifts = (shifts[0] - pcor.shape[0] / 2, shifts[1] - pcor.shape[1] / 2)
    shifts = [int(i) for i in shifts]  # convert to integers

    if shift_func == 'shift':
        # shift image using ndimage.shift
        output = ndimage.interpolation.shift(image, shifts, order=0)
    elif shift_func == 'roll':
        # shift image using roll to be reversible
        output = np.roll(image, shifts[0], axis=0)
        output = np.roll(output, shifts[1], axis=1)

    return output, shifts


def stack_align(stack, align_type='static', real_filter=1, k_filter=1, shift_func='shift'):
    """ Align a series of images by cross-correlation. All images are aligned to the first image Uses image_correlate
    which is based on simple cross correlation.

    Notes
    -----
        You should probably use ncempy.algo.stack_align since it uses mutlicorr and is more
        functional.

    Parameters
    ----------
        stack : ndarray, 3D
            The stack of images to align. Shape [num, Y, X]

        real_filter : ndarray, optional, default = 1
            A real space filter to apply before cross-correlation
            of each image. Shape must be [Y, X]

        k_filter : ndarray, optional, default = 1
            A Fourier space filter to apply before cross-correlation.
            Shape must be [Y, X/2 + 1]

        shift_func : str, default is 'shift'
            The function to use to shift the images. 'roll' uses np.roll and 'shift' uses ndimage.shift.

        align_type: str
            static or dynamic alignment. Static aligns all images to the first image. Dynamic aligns
            each image to the previous image starting with the first image

    Returns
    -------
        : tuple, aligned stack, shifts
            A tuple containing the aligned images as a 3D ndarray of shape
            [num, Y, X] and shifts as a 2D ndarray of shape [num, 2]
    """

    if align_type is not 'static' and align_type is not 'dynamic':
        raise KeyError('Incorrect align type. Must be static or dynamic')

    # Pre-allocate the arrays
    aligned = np.zeros_like(stack)  # shifted data array
    shifts = np.zeros((stack.shape[0], 2))  # the applied shifts

    aligned[0, :, :] = stack[0, :, :]

    jj = 0
    ref_sh = np.zeros((2,))
    for ii in range(1, stack.shape[0]):
        output, sh = image_correlate(stack[ii, :, :], stack[jj, :, :], real_filter, k_filter, shift_func=shift_func)
        sh += ref_sh
        aligned[ii, :, :] = output
        shifts[ii, :] = sh
        if align_type is 'dynamic':
            ref_sh = sh
            jj = ii

    return aligned, shifts



import numpy as np
from ..algo.multicorr_funcs import *


def multicorr(g1, g2, method='cross', upsample_factor=1, verbose=False):
    """Align a reference to an image by cross correlation. The template
    and the image must have the same size.

    The function takes in FFTs so that any FFT algorithm can be used to
    transform the image and template (fft2, mkl, scipack, etc.)

    Parameters
    ----------
        g1 : complex ndarray
            Fourier transform of reference image.
        g2 : complex ndarray
            Fourier transform of the image to register (the kernel).
        method : str, optional
            The correlation method to use. Must be 'phase' or 'cross' or 'hybrid' (default = 'cross')
        upsample_factor : int
            Upsample factor for subpixel precision of cross correlation. (default = 1)
        verbose : bool, default is False
            Print output.

    Returns
    -------
        xyShift : list of floats
            The shift between G1 and G2 in pixels.

    Example
    -------
        Cross correlate two images already stored as ndarrays. You must input the FFT
        of the images.

        >>> import ncempy.algo as neval
        >>> import numpy as np
        >>> im0FFT = np.fft.fft2(im0)
        >>> im1FFT = np.fft.fft2(im1)
        >>> shifts = neval.multicorr(im0FFT, im1FFT)
    """

    # Check to make sure both G1 and G2 are arrays
    if type(g1) is not np.ndarray:
        raise TypeError('G1 must be an ndarray')
    elif type(g2) is not np.ndarray:
        raise TypeError('G2 must be an ndarray')

    # Check that the inputs are complex FFTs (common error)
    if not np.iscomplexobj(g1) or not np.iscomplexobj(g2):
        raise TypeError('G1 and G2 must be complex FFTs.')

    # Check to make sure method and upsample factor are the correct values
    if method not in ['phase', 'cross', 'hybrid']:
        print('Unknown method used, setting to cross.')
        method = 'cross'

    if type(upsample_factor) is not int and type(upsample_factor) is not float:
        print('Upsample factor is not an integer or float, setting to 1')
        upsample_factor = 1
    elif type(upsample_factor) is not int:
        print('Upsample factor is not an integer, rounding down')
        upsample_factor = int(upsample_factor)
        if upsample_factor < 1:
            print('Upsample factor is < 1, setting to 1')
            upsample_factor = 1

    if upsample_factor < 1:
        raise ValueError('upsample_factor must be >= 1')

    if verbose:
        print('upsample factor = {}'.format(upsample_factor))

    # Verify images are the same size.
    if g1.shape != g2.shape:
        raise TypeError('G1 and G2 are not the same size, G1 is {0} and G2 is {1}'.format(g1.shape, g2.shape))

    imageCorr = initial_correlation_image(g1, g2, method, verbose=verbose)
    xyShift = upsampled_correlation(imageCorr, upsample_factor, verbose=verbose)

    return xyShift



"""
Module to correlate two images, functionally written.

TODO
----
    - Cant use rfft2 currently. This gives one shift as 1/2 the value. How
      can this be improved to improve speed?
"""

import numpy as np


def initial_correlation_image(g1, g2, method='cross', verbose=False):
    """Generate correlation image at initial resolution using the method specified.

    Parameters
    ----------
        g1 : complex ndarray
            Fourier transform of reference image.
        g2 : complex ndarray
            Fourier transform of the image to register (the kernel).
        method : str, optional
            The correlation method to use. Must be 'phase' or 'cross' or 'hybrid' (default = 'cross')
        verbose : bool, default is False
            Print output.
    Returns
    -------
        imageCorr : ndarray complex
            Correlation array which has not yet been inverse Fourier transformed.
    """
    if verbose:
        print('Method is {}'.format(method))
    G12 = g2 * np.conj(g1)  # is this the correct order that we want?
    if method == 'phase':
        imageCorr = np.exp(1j * np.angle(G12))
    elif method == 'cross':
        imageCorr = G12
    elif method == 'hybrid':
        imageCorr = np.sqrt(np.abs(G12)) * np.exp(1j * np.angle(G12))
    else:
        raise TypeError('{} method is not allowed'.format(str(method)))

    return imageCorr


def upsampled_correlation(image_corr, upsample_factor, verbose=False):
    """Upsamples the correlation image by a set integer factor upsample_factor.
    If upsample_factor == 2, then it is naively Fourier upsampled.
    If the upsample_factor is higher than 2, then it uses dftUpsample, which is
    a more efficient way to Fourier upsample the image.

    Parameters
    ----------
        image_corr : ndarray complex
            Fourier transformed correlation image returned by initial_correlation_image.
        upsample_factor : int
            Upsampling factor.
        verbose : bool
            Provide output for debugging

    Returns
    -------
        xyShift : list
            Shift in x and y of G2 with respect to G1.
    """
    imageCorrIFT = np.real(np.fft.ifft2(image_corr))
    xyShift = list(np.unravel_index(imageCorrIFT.argmax(), imageCorrIFT.shape, 'C'))
    if verbose:
        print('xyShift initial = {}'.format(xyShift))
    if upsample_factor == 1:
        imageSize = imageCorrIFT.shape
        xyShift[0] = ((xyShift[0] + imageSize[0]/2) % imageSize[0]) - imageSize[0]/2
        xyShift[1] = ((xyShift[1] + imageSize[1]/2) % imageSize[1]) - imageSize[1]/2
    else:
        imageCorrLarge = upsampleFFT(image_corr, 2)
        imageSizeLarge = imageCorrLarge.shape
        xySubShift2 = list(np.unravel_index(imageCorrLarge.argmax(), imageSizeLarge, 'C'))
        if verbose:
            print('xySubShift2 = {}'.format(xySubShift2))
        xySubShift2[0] = ((xySubShift2[0] + imageSizeLarge[0]/2) % imageSizeLarge[0]) - imageSizeLarge[0]/2
        xySubShift2[1] = ((xySubShift2[1] + imageSizeLarge[1]/2) % imageSizeLarge[1]) - imageSizeLarge[1]/2
        xyShift = [i/2 for i in xySubShift2]  # signs have to flip, or mod wrong?
        if verbose:
            print('xyShift line 127 = {}'.format(xyShift))

        if upsample_factor > 2:
            # here is where we use DFT registration to make things much faster
            # we cut out and upsample a peak 1.5 by 1.5 px from our original correlation image.

            xyShift[0] = np.round(xyShift[0] * upsample_factor) / upsample_factor
            xyShift[1] = np.round(xyShift[1] * upsample_factor) / upsample_factor

            globalShift = np.fix(np.ceil(upsample_factor * 1.5) / 2)
            if verbose:
                print('globalShift = {}'.format(globalShift))
                print('xyShift = {}'.format(xyShift))

            imageCorrUpsample = np.conj(dftUpsample(np.conj(image_corr), upsample_factor,
                                                    globalShift - np.multiply(xyShift, upsample_factor))) / (np.fix(imageSizeLarge[0]) * np.fix(imageSizeLarge[1]) * upsample_factor ** 2)

            xySubShift = np.unravel_index(imageCorrUpsample.argmax(), imageCorrUpsample.shape, 'C')
            if verbose:
                print('xySubShift = {}'.format(xySubShift))

            # add a subpixel shift via parabolic fitting
            try:
                icc = np.real(imageCorrUpsample[xySubShift[0] - 1 : xySubShift[0] + 2,
                                                xySubShift[1] - 1 : xySubShift[1] + 2])
                dx = (icc[2, 1] - icc[0, 1]) / (4 * icc[1, 1] - 2 * icc[2, 1] - 2 * icc[0, 1])
                dy = (icc[1, 2] - icc[1, 0]) / (4 * icc[1, 1] - 2 * icc[1, 2] - 2 * icc[1, 0])
            except:
                dx, dy = 0, 0 #  peak is near the edge and one of the above values does not exist

            xySubShift = xySubShift - globalShift
            xyShift = xyShift + (xySubShift + np.array([dx, dy])) / upsample_factor

    return xyShift


def upsampleFFT(image_init, upsample_factor):
    """This does a Fourier upsample of the imageInit. imageInit is the Fourier transform of the correlation image.
    The function returns the real space correlation image that has been Fourier upsampled by the upsample_factor.
    An upsample factor of 2 is generally sufficient.

    The way it works is that it embeds imageInit in a larger array of zeros, then does the inverse Fourier transform to return the Fourier upsampled image in real space.

    Parameters
    ----------
        image_init : ndarray  complex
            The image to be Fourier upsampled. This should be in the Fourier domain.
        upsample_factor : int
            THe upsample factor (usually 2).

    Returns
    -------
        imageUpsampleReal : ndarray complex
            The inverse Fourier transform of imageInit upsampled by the upsampleFactor.
    OLD
    ---
    imageSize = imageInit.shape
    imageUpsample = np.zeros(tuple((i*upsampleFactor for i in imageSize))) + 0j
    imageUpsample[:imageSize[0], :imageSize[1]] = imageInit
    imageUpsample = np.roll(np.roll(imageUpsample, -int(imageSize[0]/2), 0), -int(imageSize[1]/2),1)
    imageUpsampleReal = np.real(np.fft.ifft2(imageUpsample))
    return imageUpsampleReal
    """
    
    ss = [int(ii * upsample_factor / 4) for ii in image_init.shape]  # pad size
    imageInit2 = np.pad(np.fft.fftshift(image_init), ss, mode='constant')  # pad the FFT
    image_upsample_real = np.real(np.fft.ifftn(np.fft.ifftshift(imageInit2)))  # inverse FFT

    return image_upsample_real


def dftUpsample(image_corr, upsample_factor, xy_shift):
    """
    This performs a matrix multiply DFT around a small neighboring region of the initial correlation peak.
    By using the matrix multiply DFT to do the Fourier upsampling, the efficiency is greatly improved.
    This is adapted from the subfunction dftups found in the dftregistration function on the Matlab File Exchange.

    https://www.mathworks.com/matlabcentral/fileexchange/18401-efficient-subpixel-image-registration-by-cross-correlation

    The matrix multiplication DFT is from
    Manuel Guizar-Sicairos, Samuel T. Thurman, and James R. Fienup, "Efficient subpixel image registration algorithms,"
    Opt. Lett. 33, 156-158 (2008). http://www.sciencedirect.com/science/article/pii/S0045790612000778

    Parameters
    ----------
        image_corr : ndarray
            Correlation image between two images in Fourier space.
        upsample_factor : int
            Scalar integer of how much to upsample.
        xy_shift : list of 2 floats
            Single pixel shift between images previously computed. Used to center the matrix multiplication
            on the correlation peak.

    Returns
    -------
        image_upsample : ndarray
            Upsampled image from region around correlation peak.

    """
    imageSize = image_corr.shape
    pixelRadius = 1.5
    numRow = np.ceil(pixelRadius * upsample_factor)
    numCol = numRow

    colKern = np.exp((-1j * 2 * np.pi / (imageSize[1] * upsample_factor)) *
                     (np.fft.ifftshift( (np.arange(imageSize[1]))) -
                     np.floor(imageSize[1]/2)) *
                     (np.arange(numCol) - xy_shift[1])[:, np.newaxis]
                     )

    rowKern = np.exp(
    (-1j * 2 * np.pi / (imageSize[0] * upsample_factor))
    * (np.arange(numRow) - xy_shift[0])
    * (np.fft.ifftshift(np.arange(imageSize[0]))
    - np.floor(imageSize[0]/2))[:, np.newaxis]
    ) # Comment from above applies.

    image_upsample = np.real(np.dot(np.dot(rowKern.transpose(), image_corr), colKern.transpose()))

    return image_upsample


def imageShifter(g1, xy_shift):
    """
    Multiply im by a plane wave that has the real space effect of shifting ifft2(G2) by [x, y] pixels.

    Parameters
   -----------
        g1 : complex ndarray
            The Fourier transform of an image.
        xy_shift : list
            A two element list of the shifts along each axis.

    Returns
    -------
        G2shift : complex ndarray
            Fourier shifted image FFT

    Example
    -------

        >>> shiftIm0 = np.real(np.fft.ifft2(multicorr.imageShifter(np.fft.fft2(im0),[11.1,22.2])))
        >>> plt.imshow(shiftIm0)

    """
    # Check that the inputs are complex FFTs (common error)
    if not np.iscomplexobj(g1):
        raise TypeError('g1 must be complex FFTs.')

    imageSize = g1.shape
    qx = np.fft.fftfreq(imageSize[0], 1)  # does this need to be a column vector
    if imageSize[1] == imageSize[0]:
        qy = qx
    else:
        qy = np.fft.fftfreq(imageSize[1], 1)

    G2shift = np.multiply(g1, np.outer(np.exp(-2j * np.pi * qx * xy_shift[0]), np.exp(-2j * np.pi * qy * xy_shift[1])))

    return G2shift



"""
Module to find local maxima in an image.
"""

import numpy as np
import scipy.ndimage.filters


def local_max(img, r, thresh):
    """Find local maxima from comparing dilated and eroded images.

    Calculates images with maximum and minimum within given radius. If the difference is larger than the threshold, the
    original pixel position with max value is detected as local maximum.

    Parameters
    ----------
        img : np.ndarray
            Input image.
        r : int
            Radius for locality.
        thresh : int or float
            Intensity difference threshold.

    Returns
    -------
        : np.ndarray
            Array of points.

    """
    
    try:
        r = int(r)
        thresh = int(thresh)
        assert(isinstance(img, np.ndarray))
    except:
        raise TypeError('Bad input!')

    # prepare circular kernel
    y, x = np.ogrid[-r:r+1, -r:r+1]
    kernel = x**2 + y**2 <= r**2
    
    # calculate max and min images
    img_dil = scipy.ndimage.filters.maximum_filter(img, footprint=kernel)
    img_ero = scipy.ndimage.filters.minimum_filter(img, footprint=kernel)
    
    # get selection of local maxima
    sel = (img == img_dil)*(img-img_ero > thresh)
    
    if sel.any():  
        # retrieve and return points
        points = np.argwhere(sel)

        return points
    else:
        # otherwise return None to avoid having an empty list
        return None


def points_todim(points, dims):
    """Convert points from px coordinates to real dim.

    Points are expected to be array indices for the first two dimensions in dims.

    Parameters
    ----------
        points : np.ndarray
            Points to convert.
        dims : tuple
            Tuple of dimensions.

    Returns
    -------
        : np.ndarray
           Converted points.

    """
    
    try:
        # try to convert input to np.ndarray with 2 columns (necessary if only one entry provided)
        points = np.reshape(np.array(points), (-1,2))
        # check if enough dims available
        assert(len(dims) >= 2)
        assert(len(dims[0]) == 3)
    except:
        raise TypeError('Something wrong with the input!')
    
    # do the conversion by looking up thing in dimension vectors
    points_d = np.array([dims[0][0][points[:, 0]], dims[1][0][points[:, 1]]]).transpose()
    
    return points_d



import numpy as np


def shearImage(im, dim, shear_factor):
    """ Exact shear of an image using the frequency space of the FFT
    of the image.

    Currently only works for square images.

    Parameters
    ----------
        im: ndarray, 2D
            The image to shear.
        dim: int
            The axis to shear along
        shear_factor: float
            The amount of shear.

    Returns
    -------
        : ndarray
            The sheared image.
    """

    if im.shape[0] != im.shape[1]:
        print('Input image must be square.')
        return

    fftF = np.fft.fftshift(np.fft.fftfreq(im.shape[dim], d=1 / im.shape[dim]))
    yy, xx = np.meshgrid(fftF, fftF, indexing='ij')
    f = np.fft.fftshift(np.fft.fftn(im, axes=(dim,)), axes=(dim,))
    ff = f * np.exp(-2.0j * np.pi / im.shape[dim] * xx * yy * shear_factor)

    sh = np.fft.ifftn(np.fft.ifftshift(ff, axes=(dim,)), axes=(dim,))
    sh = np.real(sh)
    return sh


def shiftImage(im, shift):
    """ Exact shear of an image using the frequency space of the FFT
    of the image.

    Parameters
    ----------
        im: ndarray, 2D
            The image to shear.
        shift: tuple of floats
            The number of pixels to shift in each direction
    Returns
    -------
        : ndarray
            The shifted image.
    """

    fftF0 = np.fft.fftshift(np.fft.fftfreq(im.shape[0], d=1 / im.shape[0]))
    fftF1 = np.fft.fftshift(np.fft.fftfreq(im.shape[1], d=1 / im.shape[1]))
    xx, yy = np.meshgrid(fftF0, fftF1, indexing='ij')
    f = np.fft.fftshift(np.fft.fftn(im))
    ff = f * np.exp(-2.0j * np.pi / im.shape[0] * xx * shift[0] +
                    -2.0j * np.pi / im.shape[1] * yy * shift[1])

    sh_im = np.fft.ifftn(np.fft.ifftshift(ff))
    sh_im = np.real(sh_im)
    return sh_im


def rotateImage(im, theta, pad=False):
    """ Use three shears in Fourier space to exactly (and reversibly) rotate
    an image.

    Currently only works for square images.

    Parameters
    ----------
        im: ndarray, 2D
            The image to rotate

        theta: float
            The angle to rotate by in radians

        pad: bool
            Add padding to the image before rotating

    """

    if im.shape[0] != im.shape[1]:
        print('Input image must be square.')
        return

    alpha = gamma = - np.tan(theta/2.)
    beta = np.sin(theta)

    if pad:
        p = im.shape[0] * np.sin(theta)
        im = np.pad(im, int(p//2+1))

    im1 = shearImage(im,  0, alpha)
    im2 = shearImage(im1, 1, beta)
    im3 = shearImage(im2, 0, gamma)
    im3 = np.real(im3)
    return im3



# import genfire as gf
from genfire import reconstruct, fileio
import sys, os, shutil
import argparse

# Parse command line arguments.
parser = argparse.ArgumentParser()
parser.add_argument("fileroot", help="There should be an aligned tomo stack named fileroot_aligned.npy in the directory and a the reconstruction will be named fileroot_reconstruction.mrc.")
parser.add_argument("-n", "--numIterations", help="Default number of GENFIRE iterations.  See GENFIRE documentation.  Default=100.", type=int, default=100, action="store")
parser.add_argument("-o", "--oversamplingRatio", help="How many times bigger to make the fourier space.  Bigger is more accurate, but you may run out of memory.  See GENFIRE documentation.  Default=4.", type=int, default=4, action="store")
args = parser.parse_args()

# Make a reconstruction using paramaters from default or from the command line.
GF = reconstruct.GenfireReconstructor(
        projections = args.fileroot+"_aligned.npy",
        eulerAngles = "tilts.txt",
        resultsFilename = args.fileroot+"_reconstruction.mrc",
        numIterations = args.numIterations,
        interpolationCutoffDistance=0.7,
        oversamplingRatio=args.oversamplingRatio,
        resolutionExtensionSuppressionState=2,
        )

# Number crunch.
results = GF.reconstruct()

# Save results.
fileio.saveResults(results, GF.params.resultsFilename)



"""
Module containing definitions of basic math functions, which can be used for fitting.

If you add functions here, do not forget to update the lookup table at the bottom as well.
"""

import numpy as np
import scipy.special


def const( x, param):
    """Constant function.

    f = param[0] .

    Parameters:
        x (np.ndarray):    Positions at which to evaluate function.
        param (np.ndarray):    Necessary parameters.

    Returns:
        (np.ndarray):    Values of function at x.

    """
    
    return param[0]*np.ones(x.shape)


def linear( x, param):
    """Linear function.

    f = param[0] * x + param[1] .

    Parameters:
        x (np.ndarray):    Positions at which to evaluate function.
        param (np.ndarray):    Necessary parameters.

    Returns:
        (np.ndarray):    Values of function at x.

    """
    
    return param[0]*x + param[1]


def powlaw( x, param):
    """Power law.

    f = param[0] * x^param[1] .

    Parameters:
        x (np.ndarray):    Positions at which to evaluate function.
        param (np.ndarray):    Necessary parameters.

    Returns:
        (np.ndarray):    Values of function at x.

    """
    return param[0]*np.power(x, param[1])
    

def voigt(x, param):
    """
    Voigt peak function.

    f = param[0]/(param[2] * sqrt(2*pi)) * Re( Faddeeva( ((x - param[1]) + i*param[3])/(param[2] * sqrt(2)) ) )

    Parameters:
        x (np.ndarray):    Positions at which to evaluate function.
        param (np.ndarray):    Necessary parameters.

    Returns:
        (np.ndarray):    Values of function at x.

    """
    # A = param[0]
    # mu = param[1]
    # sigma = param[2]
    # gamma = param[3]
    return param[0]*np.real(scipy.special.wofz((x-param[1] + 1.0j*param[3])/(param[2]*np.sqrt(2.0))))/(param[2]*np.sqrt(2.0*np.pi))


def sum_functions( x, funcs, param ):
    """
    Sum functions in funcs over range x.

    Parameters:
        x (np.ndarray):    Positions at which to evaluate the functions.
        funcs (list):    List of strings identifying function implemented in ncempy.algo.math.
        param (np.ndarray):    Concatenated parameters for functions in funcs.

    Returns:
        (np.ndarray):   Values of sum of functions at x.

    """
    
    # estimator
    est = np.zeros(x.shape)
    
    n = 0
    # evaluate given functions
    for i in range(len(funcs)):
        est += lkp_funcs[funcs[i]][0]( x, param[n:n+lkp_funcs[funcs[i]][1]] )
        n += lkp_funcs[funcs[i]][1]
        
    return est
    

# lookup table for functions
lkp_funcs = {'const': (const, 1),
              'linear': (linear, 2),
              'powlaw': (powlaw, 2),
              'voigt': (voigt, 4)
            }
'''(dict):    Look-up table for functions implemented in this module. Functions are identified by strings, the entries give handles to the functions as well as the number of arguments necessary.'''



import sys, os, shutil
from collections import OrderedDict
import glob2 as glob
import numpy as np
import hyperspy.api as hs
from ncempy.io.emd import fileEMD

def GetTiltsFromBrukerSequence(Directory=None):
    ''' Find the set of Bruker bcf files and figure out their tilts based on filenames.

    There should be a set of bcf files in a directory with names: ... -10.bcf, -5.bcf, 0.bcf, 5.bcf ...
    The number represents the stage tilt of that stack.
    The angles do not matter but the names must be coded as above.  They will be automatically read and sorted.

    Parameters:
        Directory (str): Name of directory containing the bcf files.

    Returns:
        (list of floats): A list of tilt angles.
    '''

    Bcfs = glob.glob(os.path.join(Directory, '*.bcf'))
    NewBcfs = []
    for i in Bcfs:
        _ , FileName = os.path.split(i)
        NewBcfs.append(FileName)
    Tilts = list(map(lambda s: float(s[:-4]), NewBcfs))
    Tilts.sort()

    return Tilts

def GetSpatialDimension(Dim):
    ''' Return a spatial value in meters. '''
    unit = Dim.units
    if unit == 'nm':
        scalefac = 1e-9
    elif unit == 'µm':
        scalefac = 1e-6
    elif unit == 'mm':
        scalefac = 1e-3
    elif unit == 'cm':
        scalefac = 1e-2
    else:
        print('Unrecognized spatial dimension unit in Bruker file: ' + unit)
        scalefac = 1
    Axis = Dim.index2value(range(Dim.size))
    return Axis * scalefac

def GetEnergyDimension(Dim):
    ''' Return an energy value in eV.  '''
    unit = Dim.units
    if str(unit) == 'eV':
        scalefac = 1
    elif str(unit) == str('keV'):
        scalefac = 1e3
    elif unit == 'MeV':
        scalefac = 1e6
    else:
        print('Unrecognized energy dimension unit in Bruker file: ' + unit)
        scalefac = 1
    Axis = Dim.index2value(range(Dim.size))
    return Axis * scalefac

def ExtractRawSignalsFromBrukerSequence(InputDirectory=None, OutputEMD=None):
    ''' Read in a set of Bruker bcf files containing EDS acquisitions and write an EMD file with the data.

    Parameters:
        InputDirectory (str): Name of directory containing the bcf files.

        OutputEMD (str): Name of the output file (may include a path).

    Returns:
        None.
    '''

    # Find all the bcf files first.
    Tilts = GetTiltsFromBrukerSequence(Directory=InputDirectory)
    Tilts = np.array(Tilts)

    # We need some filename if none was given to us.
    if OutputEMD is None:
        OutputEMD = 'test.emd'
        
    print('Extracting Signals:')
    for i, t in enumerate(Tilts):
        # Load the bruker file for this tilt.
        fname = os.path.join(InputDirectory, str(int(t))+'.bcf')
        x = hs.load(fname)

        # Only the first time, we need to calculate the sizes of all the arrays we are going to make.
        if 'HAADFsize_m' not in locals():
            # HAADF's have x,y dimensions.
            HAADFsize_m = GetSpatialDimension(x[0].axes_manager['width'])
            HAADFsize_n = GetSpatialDimension(x[0].axes_manager['height'])
            HAADFDim = (len(Tilts), len(HAADFsize_m), len(HAADFsize_n))
            print('HAADF has dimensions ' + str(HAADFDim))

            # EDS cubes have x,y, energy dimensions.
            EDSsize_m = GetSpatialDimension(x[1].axes_manager['width'])
            EDSsize_n = GetSpatialDimension(x[1].axes_manager['height'])
            EDSsize_e = GetEnergyDimension(x[1].axes_manager['Energy'])
            EDSDim = (len(Tilts), len(EDSsize_m), len(EDSsize_n), len(EDSsize_e))
            print('EDS has dimensions ' + str(EDSDim))

            HAADF = np.zeros(HAADFDim)
            EDS = np.zeros(EDSDim, dtype='float32')

            BeamEnergy = x[0].metadata['Acquisition_instrument']['TEM']['beam_energy']*1000 # eV
            EnergyResolutionMnKa = x[1].metadata['Acquisition_instrument']['TEM']['Detector']['EDS']['energy_resolution_MnKa'] # eV
            DetectorTiltAngle = x[1].metadata['Acquisition_instrument']['TEM']['Detector']['EDS']['elevation_angle'] # degrees
            RealTime = x[1].metadata['Acquisition_instrument']['TEM']['Detector']['EDS']['real_time'] # seconds

        print(str(fname))

        HAADF[i,:,:] = x[0].data.astype("float32")
        EDS[i,:,:,:] = x[1].data.astype("float32") 

    # open nonexisting file for writing
    if os.path.isfile(OutputEMD):
        os.remove(OutputEMD)
    EMD = fileEMD(OutputEMD)

    data = HAADF
    dims = ( (Tilts, 'angle', '[deg]'),
             (HAADFsize_m, 'x', '[m]'),
             (HAADFsize_n, 'y', '[m]'))
    print('Writing HAADF tilt stack.')
    EMD.put_emdgroup('HAADF_TiltStack', data, dims)

    data = EDS
    dims = ( (Tilts, 'angle', '[deg]'),
             (EDSsize_m, 'x', '[m]'),
             (EDSsize_n, 'y', '[m]'),
             (EDSsize_e, 'E', '[eV]')) 
    print('Writing EDS tilt stack.')
    EMD.put_emdgroup('EDS_TiltStack', data, dims)

    EMD.microscope.attrs['BeamVoltage[eV]'] = BeamEnergy
    EMD.microscope.attrs['MnKaResolution[eV]'] = EnergyResolutionMnKa
    EMD.microscope.attrs['DetectorTiltAngle[deg]'] = DetectorTiltAngle
    EMD.microscope.attrs['RealTime[s]'] = RealTime * len(Tilts)

    EMD.put_comment('File created.')
    del EMD
    print('Created file ' + OutputEMD)




from . import dm
from . import ser
from . import emd
from . import mrc
from . import emdVelox
# from .read import read

from pathlib import Path


def read(filename, dsetNum=0):
    """
    Parameters
    ----------
    filename : str or pathlib.Path
        The path and name of the file to attempt to load. This chooses the Reader() function based on the file
        suffix.
    dsetNum : int
        The data set number to load if there are multiple data sets in a file. This is implemented for EMD and DM
        files only.
    """

    out = {}

    # check filename type
    if isinstance(filename, str):
        filename = Path(filename)
    elif isinstance(filename, Path):
        pass
    else:
        raise TypeError('Filename is supposed to be a string or pathlib.Path')

    if not filename.exists():
        raise FileNotFoundError

    # ensure lowercase suffix
    suffix = filename.suffix.lower()

    if suffix == '.ser':
        out = ser.serReader(filename)
    elif suffix in ('.dm3', '.dm4'):
        out = dm.dmReader(filename)
    elif suffix in ('.emd', '.h5', '.hdf5'):
        is_velox = False
        with emd.fileEMD(filename) as emd0:
            if len(emd0.list_emds) > 0:
                out = emd.emdReader(filename, dsetNum)
            else:
                is_velox = True
        if is_velox:
            out = emdVelox.emdVeloxReader(filename, dsetNum)
    elif suffix in ('.mrc', '.rec', '.st', '.ali'):
        out = mrc.mrcReader(filename)
    else:
        print('File suffix {} is not recognized.'.format(suffix))
        print('Supported formats are ser, mrc, rec, ali, st, emd, dm3, dm4.')

    return out



from io import StringIO
import numpy as np
import pickle
import itertools
import os, sys, shutil

ElamLoaded = False # We only want to load the Elam database once.  Keep track of if that has been done.  False = not yet loaded.
ElamData = []

# For debugging we sometimes want Elam to print out the lines it loaded.
ElamPrint = lambda s: None #print(s)

# Reading the Elam database is time consuming.  This caches the fluorescence data so it is only read once per program
#  load.
ElamFluorescenceLines = {'nodata':1}

def GetElamFluorescenceLines(ElementName, E=None, I=None):
    global ElamLoaded, ElamData, ElamFluorescenceLines

    # Already loaded fluorescence data for this element?
    if ElementName in ElamFluorescenceLines:
        # Return the (previously or newly) cached element info.
        return ElamFluorescenceLines[ElementName]

    # This is really time consuming, so we also cache it as a pickle file.
    try:
        ElementXRayLines = pickle.load(open(os.path.join(os.path.dirname(__file__), 'Elam', ElementName.lower() +'.pickle'), 'rb'))
        # It was pickled!  So cache it in memory and return it.
        ElamFluorescenceLines[ElementName] = ElementXRayLines
        return ElementXRayLines
    except:
        # It wasn't pickled, no matter.  We'll make it now and pickle it at the end of the function.
        pass

    # It's not already loaded, so continue with loading it.

    # Read the entire Elam database into memory.
    if ElamLoaded == False:
        fo = open(os.path.join(os.path.dirname(__file__), 'Elam', 'ElamDB12.txt'), "r+")
        str = fo.read();
        fo.close()
        ElamData = StringIO(str)
        ElamLoaded = True

    # Initialize the lists that will contain the line information.
    LineNamesIUPAC = []
    LineNamesSiegbahn = []
    Geoms = []

    # Make sure we're at the start of the file
    ElamData.seek(0)

    # Zoom forward to the element we are looking for.
    for line in ElamData:
        if ('Element '+ElementName+' ') in line:
            break

    # Pull out the section with the fluorescence lines, which is first.
    for line in ElamData:
        # Ignore the lines that aren't the fluorescence info
        if ('Edge' in line) or ('Lines' in line) or ('CK' in line):
            continue
        # When we get to the Photoionization section we are done.
        if 'Photo' in line:
            break;
        # Otherwise, this is a fluorescence line.  Add it to our output if it is intense enough to improve the fit
        q = line.split()
        if (float(q[3]) > 0.01):
            LineNamesIUPAC.append(q[0])
            LineNamesSiegbahn.append(q[1])
            Geoms.append([float(q[2]), float(q[3])])
            ElamPrint('Added: ' + line)
        else:
            ElamPrint('Ignored (intensity < 1%): ' + line)

    Geoms = np.array(Geoms)

    # We are going to use the Siegbahn names, not IUPAC.
    # Make a dictionary where the key is the line name and the value is an (energy, intensity) tuple.
    Lines = dict(zip(LineNamesSiegbahn, Geoms))

    ElementXRayLines = {}

    for Series in ['K', 'L', 'M', 'N', 'O']:

        #Obtain the name strings for only the lines that have names that start with K (or L, or M, whatever Series we are on).
        ThisSeries = dict([(k, v) for k, v in Lines.items() if k.startswith(Series)])

        # If these series is not present, then don't keep it.
        if len(ThisSeries) == 0:
            continue

        # Make this edge in ElementXRayLines
        ElementXRayLines[Series] = ThisSeries

    # We've built the fluorescence data for this element.  Cache it and pickle it for next time.
    ElamFluorescenceLines[ElementName] = ElementXRayLines
    pickle.dump(ElementXRayLines, open(os.path.join(os.path.dirname(__file__), 'Elam', ElementName.lower() + '.pickle'), 'wb'))

    # And return what we've got.
    return ElementXRayLines

def GetWeightedSum(Line, Series):
    # The key is that the series we want to sum has to have a common prefix.
    # For example, if Line = 'Ka' then we can sum 'Ka1', 'Ka2' but not 'Kbx'
    # Get only the lines that match our specific subseries.
    vals = np.array([Vals for LineName, Vals in Series.items() if Line in LineName])
    # Line weights are summed to one for the entire series.  Since we often want to weigh a subseries, renormalize.
    vals[:,1] /= np.sum(vals[:,1])
    # Now get a weighted sum: sum(position*relative_intensity)
    WeightedEnergy = np.sum(vals[:,0]*vals[:,1])
    return WeightedEnergy

def GetFluorescenceLineEnergy(ElementName, Series='K', Line=None):
    LineData = GetElamFluorescenceLines(ElementName)

    # If the user requests a "line" which is the series, we will return the weighted sum of all lines in the series (below in the if).
    if Line == Series:
        Line = None

    # If the user specifies a specific line, then we will return its energy.
    try:
        if Line in ['Ka', 'Kb', 'La', 'Lb']:
            # We can make weighted sums of special subseries of lines. 
            return GetWeightedSum(Line, LineData[Series])
        elif Line is not None:
            # Or we can return a specific line which is requested.
            # print(LineData)
            return LineData[Series][Line][0]
        else:
            # The user doesn't specify a line which means he wants to know the energy of the whole series.
            return GetWeightedSum(Series, LineData[Series])
    except:
        # If the user asked for a series or line that's not in the database, then we return none.
        return None

if __name__ == '__main__':
    print('------------------------------')
    # print('Fe-K: %s' % GetFluorescenceLineEnergy('Fe'))
    # print('Al-K: %s' % GetFluorescenceLineEnergy('Al'))
    # print('Mg-K: %s' % GetFluorescenceLineEnergy('Mg'))
    # print('Mg-K: %s' % GetFluorescenceLineEnergy('Mg', Series='K'))
    # print('Mg-K: %s' % GetFluorescenceLineEnergy('Mg', Series='K', Line='K'))
    # print('Fe-Q: %s' % GetFluorescenceLineEnergy('Fe', Series='Q'))
    # print('Fe-Ka2: %f' % GetFluorescenceLineEnergy('Fe', Series='K', Line='Ka2'))
    # print('Fe-Ka1: %f' % GetFluorescenceLineEnergy('Fe', Series='K', Line='Ka1'))
    # print('Fe-Ka: %f' % GetFluorescenceLineEnergy('Fe', Series='K', Line='Ka'))
    # print('Fe-Kb: %f' % GetFluorescenceLineEnergy('Fe', Series='K', Line='Kb'))
    # print('Mn-Ka: %f' % GetFluorescenceLineEnergy('Mn', Series='K', Line='Ka'))
    # print('Pt-La: %f' % GetFluorescenceLineEnergy('Pt', Series='L', Line='La'))
    print('S-K: %f' % GetFluorescenceLineEnergy('S', Series='K', Line='K'))



import numpy as np


def moments(im, order=3):
    """ Calculate the image raw moments.

    Parameters
    ----------
        im : ndarray
            A 2D array of the image to calculate the moments for

        order : int, default=3
            The order to calculate the moments to.

    Returns
    -------
        : ndarray
            An ndarray of (order+1, order+1)
    """
    order = int(order) + 1
    XX, YY = np.mgrid[0:im.shape[0], 0:im.shape[1]]
    mm = np.zeros((order, order))
    for ii in range(order):
        XXii = XX ** ii
        for jj in range(order):
            mm[ii, jj] = np.sum(XXii * YY ** jj * im.astype(np.float64))
    return mm


def moments_central(im, cent=None, order=3):
    """ Calculate the image central moments.

    Parameters
    ----------
        im : ndarray
            A 2D array of the image to calculate the central moments for

        cent : tuple, 2 values, optional
            The centroid calculated from the image raw moments. If it
            is not supplied then it is calculated by moments()

        order : int, default = 3
            The order to calculate the orders to.

    Returns
    -------
        : ndarray
            An ndarray of size (order + 1, order +1)

    """
    if not cent:
        m = moments(im, order)
        cent = centroid(m)

    order = int(order) + 1

    XX, YY = np.mgrid[-cent[0]:im.shape[0] - cent[0],
                      - cent[1]:im.shape[1] - cent[1]]
    XX = XX.astype('float64')
    YY = YY.astype('float64')

    mc = np.zeros((order, order))
    for ii in range(order):
        XXii = XX ** ii
        for jj in range(order):
            mc[ii, jj] = np.sum(XXii * YY ** jj * im.astype(np.float64))
    return mc


def centroid(m):
    """ Find the centroid based on the raw moments.

    Parameters
    ----------
        m : ndarray
            The output from moments()

    Returns
    -------
        : tuple, 2 elements (X,Y)
            The position of the centroid as a 2-tuple
    """
    return m[1, 0] / m[0, 0], m[0, 1] / m[0, 0]


def moment_angle(mc):
    """ Calculate the orientation of the object from its central moments

    Parameters
    ----------
        mc : ndarray
            The array of the central moments. Output from moments_central

    Returns
    -------
        : float
            The angle in radians.

    """
    mu_20 = mc[2, 0] / mc[0, 0]  # - centroid[0]**2
    mu_02 = mc[0, 2] / mc[0, 0]  # - centroid[1]**2
    mu_11 = mc[1, 1] / mc[0, 0]  # - centroid[0]*centroid[1]

    th = 0.5 * np.arctan2(2 * mu_11, mu_20 - mu_02)
    return th



"""
Module to handle distortions in diffraction patterns.
"""

import numpy as np
import scipy.optimize


def filter_ring(points, center, rminmax):
    """Filter points to be in a certain radial distance range from center.

    Parameters
    ----------
        points : np.ndarray
            Candidate points.
        center : np.ndarray or tuple
            Center position.
        rminmax : tuple
            Tuple of min and max radial distance.

    Returns
    -------
        : np.ndarray
            List of filtered points, two column array.

    """
    
    try:
        # points have to be 2D array with 2 columns
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        assert(len(points.shape) == 2)    
        
        # center can either be tuple or np.array    
        center = np.array(center)
        center = np.reshape(center, 2)
        
        rminmax = np.array(rminmax)
        rminmax = np.reshape(rminmax, 2)
   
    except:
        raise TypeError('Something wrong with the input!')
    
    # calculate radii
    rs = np.sqrt( np.square(points[:,0]-center[0]) + np.square(points[:,1]-center[1]) )
    
    # filter by given limits
    sel = (rs>=rminmax[0])*(rs<=rminmax[1])
    
    if sel.any():
        return points[sel]
    else:
        return None
    
    
def points_topolar(points, center):
    """Convert points to polar coordinate system.

    Can be either in pixel or real dim, but should be the same for points and center.

    Parameters
    ----------
        points : np.ndarray
            Positions as two column array.
        center : np.ndarray or tuple
            Origin of the polar coordinate system.

    Returns
    -------
        : np.ndarray
            Positions in polar coordinate system as two column array (r, theta).

    """
    
    try:
        # points have to be 2D array with 2 columns
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        assert(len(points.shape) == 2)    
        
        # center can either be tuple or np.array    
        center = np.array(center)
        center = np.reshape(center, 2)
   
    except:
        raise TypeError('Something wrong with the input!')
    
    # calculate radii
    rs = np.sqrt( np.square(points[:,0]-center[0]) + np.square(points[:,1]-center[1]) )
    # calculate angle
    thes = np.arctan2(points[:,1]-center[1], points[:,0]-center[0])
    
    return np.array( [rs, thes] ).transpose()


def residuals_center( param, data):
    """Residual function for minimizing the deviations from the mean radial distance.

    Parameters
    ----------
        param : np.ndarray
            The center to optimize.
        data : np.ndarray
            The points in x,y coordinates of the original image.

    Returns
    -------
        : np.ndarray
           Residuals.

    """
    
    # manually calculating the radii, as we do not need the thetas
    rs = np.sqrt( np.square(data[:,0]-param[0]) + np.square(data[:,1]-param[1]) )
    
    return rs-np.mean(rs)
    
    
def optimize_center(points, center, maxfev=1000, verbose=None):
    """Optimize the center by minimizing the sum of square deviations from the mean radial distance.

    Parameters
    ----------
        points : np.ndarray
            The points to which the optimization is done (x,y coords in org image).
        center : np.ndarray or tuple
            Initial center guess.
        maxfev : int
            Max number of iterations forwarded to scipy.optimize.leastsq().
        verbose : bool
            Set to get verbose output.

    Returns
    -------
        : np.ndarray
            The optimized center.

    """

    try:
        # points have to be 2D array with 2 columns
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        assert(len(points.shape) == 2)    
        
        # center can either be tuple or np.array    
        center = np.array(center)
        center = np.reshape(center, 2)
   
    except:
        raise TypeError('Something wrong with the input!')

    # run the optimization
    popt, flag = scipy.optimize.leastsq(residuals_center, center, args=points, maxfev=maxfev)
    
    if flag not in [1, 2, 3, 4]:
        print('WARNING: center optimization failed.')

    if verbose:
        print('optimized center: ({}, {})'.format(center[0], center[1]))

    return popt
    

def rad_dis(theta, alpha, beta, order=2):
    """Radial distortion due to ellipticity or higher order distortion.

    Relative distortion, to be multiplied with radial distance.

    Parameters
    ----------
        theta : np.ndarray
            Angles at which to evaluate. Must be float.
        alpha : float
            Orientation of major axis.
        beta : float
            Strength of distortion (beta = (1-r_min/r_max)/(1+r_min/r_max).
        order : int
            Order of distortion.

    Returns
    -------
        : np.ndarray
            Distortion factor.
    """
    
    return (1.-np.square(beta))/np.sqrt(1.+np.square(beta)-2.*beta*np.cos(order*(theta+alpha)))

    
def residuals_dis(param, points, ns):
    """Residual function for distortions.

    Parameters
    ----------
        param : np.ndarray
            Parameters for distortion.
        points : np.ndarray
            Points to fit to.
        ns : tuple
            List of orders to account for.

    Returns
    -------
        : np.ndarray
           Residuals.
    """

    est = param[0]*np.ones(points[:, 1].shape)
    for i in range(len(ns)):
        est *=rad_dis(points[:, 1], param[i*2+1], param[i*2+2], ns[i])
            
    return points[:, 0] - est
    
    
def optimize_distortion(points, ns, maxfev=1000, verbose=False):
    """Optimize distortions.

    The orders in the list ns are first fitted subsequently and the result is refined in a final fit simultaneously fitting all orders.

    Parameters
    ----------
        points : np.ndarray
            Points to optimize to (in polar coords).
        ns : tuple
            List of orders to correct for.
        maxfev : int
            Max number of iterations forwarded to scipy.optimize.leastsq().
        verbose : bool
            Set for verbose output.

    Returns
    -------
        : np.ndarray
            Optimized parameters according to ns.
    """
    
    try:
        assert(isinstance(points, np.ndarray))
        assert(points.shape[1] == 2)
        # check points to be sufficient for fitting
        assert(points.shape[0] >= 3)
        
        # check orders
        assert(len(ns)>=1)
    except:
        raise TypeError('Something wrong with the input!')

    # init guess for full fit
    init_guess = np.ones(len(ns)*2+1)
    init_guess[0] = np.mean(points[:,0])
    
    # make a temporary copy
    points_tmp = np.copy(points)
    
    if verbose:
        print('correction for {} order distortions.'.format(ns))
        print('starting with subsequent fitting:')
    
    # subsequently fit the orders
    for i in range(len(ns)):
        # optimize order to points_tmp
        popt, flag = scipy.optimize.leastsq(residuals_dis, np.array((init_guess[0], 0.1, 0.1)),
                                                     args=(points_tmp, (ns[i],)), maxfev=maxfev)
        
        if flag not in [1, 2, 3, 4]:
            print('WARNING: optimization of distortions failed.')
        
        # information
        if verbose:
            print('fitted order {}: R={} alpha={} beta={}'.format(ns[i], popt[0], popt[1], popt[2]))
        
        # save for full fit
        init_guess[i*2+1] = popt[1]
        init_guess[i*2+2] = popt[2]
        
        # do correction
        points_tmp[:, 0] /= rad_dis(points_tmp[:, 1], popt[1], popt[2], ns[i])
    
    # full fit    
    if verbose:
        print('starting the full fit:')    
    
    popt, flag = scipy.optimize.leastsq(residuals_dis, init_guess, args=(points, ns), maxfev=maxfev)
    
    if flag not in [1, 2, 3, 4]:
        print('WARNING: optimization of distortions failed.')
    
    if verbose:
        print('fitted to: R={}'.format(popt[0]))
        for i in range(len(ns)):
            print('.. order={}, alpha={}, beta={}'.format(ns[i], popt[i*2+1], popt[i*2+2]))

    return popt



from scipy import ndimage
import numpy as np


def _gen_line(p0, p1, num_points):
    """ Generate a set of points on a line between two points.
    Helper function for line_profile.


    Parameters
    ----------
        p0 : 2-tuple
            The x0 and y0 starting points of the line

        p1 : 2-tuple
            The x1 and y1 ending points of the line

        num_points : integer
            The number of points along the line

    Returns
    -------
        : tuple, x and y
            The x coordintes at which the line proifle was taken at and y the intensity values.

    """

    x = np.linspace(p0[0], p1[0], num_points)
    m = (p1[1] - p0[1]) / (p1[0] - p0[0])
    b = p0[1] - p0[0] * m
    y = m * x + b
    return x, y


def line_profile(im0, p0, p1, num_points, width=0, step=0.5):
    """ Use interpolation to measure a line scan across a 2D image. map_coordinates
    uses a different convention from numpy indexing. So the x and y positions need
    to be flipped.

    Parameters
    ----------
        im0 : ndarray
            The 2D array of the image
        p0 : tuple
            The start of the line scan (x0,y0) and (row,col)
        p1 : tuple
            The end of the line scan (x1,y1)
        num_points : int
            The number of points in the line scan
        width : int
            The width of the line scan to average over. Must be an integer.
        step : float
            The step size laterally for the interpolation. Must be < 1

    Returns
    -------
        : tuple
            A tuple containing the line profile and the positions where the interpolation was performed

    Note
    ----
        As an example: if width = 1 and step = 0.5 then there will be 3 line scans averaged.

    Example
    --------
        >>> line, (xx, yy) = line_profile(image,(0,100),(175,100),50,step=0.5,width=5)
        >>> fg, ax = plt.subplots(1,2)
        >>> ax[0].plot(line,'*-')
        >>> ax[1].imshow(image)
        >>> ax[1].scatter(xx, yy)
    """

    im = im0.astype(np.float32)

    # The line scan from p0 to p1
    (x0, y0) = _gen_line(p0, p1, num_points)
    line0 = ndimage.map_coordinates(im, np.vstack((x0, y0)))  # reverse x and y to switch between row and column major

    # Average perpendicular line scans
    if width >= 1:
        x1 = []  # save all x positions
        y1 = []  # save all y positions
        x1.append(x0)
        y1.append(y0)
        x0 = x1
        y0 = y1

        # Find perpendicular direction to line scan for averaging
        v0 = np.matrix([p1[0] - p0[0], p1[1] - p0[1], 0])
        v0 = v0 / np.linalg.norm(v0)
        v2 = np.cross(v0, np.matrix([0, 0, 1]))
        v2 = v2 / np.linalg.norm(v2)
        v2 = v2[0]  # the perpendicular direction

        # Positive direction
        numLines = int(width / step / 2.0)
        for ii in range(0, numLines):
            p2 = np.array(p0) + v2[0:2] * step * ii
            p3 = np.array(p1) + v2[0:2] * step * ii
            (x, y) = _gen_line(p2, p3, num_points)
            line0 += ndimage.map_coordinates(im, np.vstack(
                (x, y)))  # reverse x and y to switch between row and column major
            x0.append(x)
            y0.append(y)
        # Negative direction
        for ii in range(0, numLines):
            p4 = np.array(p0) - v2[0:2] * step * ii
            p5 = np.array(p1) - v2[0:2] * step * ii
            (x, y) = _gen_line(p4, p5, num_points)
            line0 += ndimage.map_coordinates(im, np.vstack(
                (x, y)))  # reverse x and y to switch between row and column major
            x0.append(x)
            y0.append(y)
    line0 = line0 / (width / step + 1)

    return line0, (x0, y0)


if __name__ == '__main__':

    XX, YY = np.mgrid[0:100,0:100]
    RR = np.sqrt((XX-50)**2 + (YY-50)**2)

    profile, (xx, yy) = line_profile(RR, (0,0), (99, 99), 100)
    print(profile)
